# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-locals
# pylint: disable=line-too-long

from typing import Dict, List, NamedTuple, Optional, Union
from msrest.serialization import Model
from azure.core.credentials import AzureSasCredential, TokenCredential


class BlobStoreInfo(NamedTuple):
    name: str
    account_name: str
    endpoint: str
    container_name: str
    credential: Optional[Union[AzureSasCredential, TokenCredential, str]]


class WorkspaceHubConfig(Model):
    """WorkspaceHub's configuration object."""

    _attribute_map = {
        "additional_workspace_storage_accounts": {"key": "additionalWorkspaceStorageAccounts", "type": "[str]"},
        "default_workspace_resource_group": {"key": "defaultWorkspaceResourceGroup", "type": "str"},
    }

    def __init__(
        self,
        *,
        additional_workspace_storage_accounts: Optional[List[str]] = None,
        default_workspace_resource_group: Optional[str] = None,
        **kwargs
    ):
        super(WorkspaceHubConfig, self).__init__(**kwargs)
        self.additional_workspace_storage_accounts = additional_workspace_storage_accounts
        self.default_workspace_resource_group = default_workspace_resource_group


class Workspace(Model):
    """An object that represents a machine learning workspace.

    Variables are only populated by the server, and will be ignored when sending a request."""

    _validation = {
        "id": {"readonly": True},
        "name": {"readonly": True},
        "type": {"readonly": True},
        #'system_data': {'readonly': True},
        "agents_endpoint_uri": {"readonly": True},
        "ml_flow_tracking_uri": {"readonly": True},
        #'notebook_info': {'readonly': True},
        # "private_endpoint_connections": {"readonly": True},
        #'private_link_count': {'readonly': True},
        "provisioning_state": {"readonly": True},
        "service_provisioned_resource_group": {"readonly": True},
        "storage_hns_enabled": {"readonly": True},
        "tenant_id": {"readonly": True},
        "workspace_id": {"readonly": True},
    }

    _attribute_map = {
        "id": {"key": "id", "type": "str"},
        "name": {"key": "name", "type": "str"},
        "type": {"key": "type", "type": "str"},
        #'system_data': {'key': 'systemData', 'type': 'SystemData'},
        #'identity': {'key': 'identity', 'type': 'ManagedServiceIdentity'},
        "kind": {"key": "kind", "type": "str"},
        "location": {"key": "location", "type": "str"},
        #'sku': {'key': 'sku', 'type': 'Sku'},
        "tags": {"key": "tags", "type": "{str}"},
        "agents_endpoint_uri": {"key": "properties.agentsEndpointUri", "type": "str"},
        "allow_public_access_when_behind_vnet": {"key": "properties.allowPublicAccessWhenBehindVnet", "type": "bool"},
        "allow_role_assignment_on_rg": {"key": "properties.allowRoleAssignmentOnRG", "type": "bool"},
        "application_insights": {"key": "properties.applicationInsights", "type": "str"},
        "associated_workspaces": {"key": "properties.associatedWorkspaces", "type": "[str]"},
        "container_registries": {"key": "properties.containerRegistries", "type": "[str]"},
        "container_registry": {"key": "properties.containerRegistry", "type": "str"},
        "description": {"key": "properties.description", "type": "str"},
        "discovery_url": {"key": "properties.discoveryUrl", "type": "str"},
        "enable_data_isolation": {"key": "properties.enableDataIsolation", "type": "bool"},
        "enable_service_side_cmk_encryption": {"key": "properties.enableServiceSideCMKEncryption", "type": "bool"},
        "enable_simplified_cmk": {"key": "properties.enableSimplifiedCmk", "type": "bool"},
        "enable_software_bill_of_materials": {"key": "properties.enableSoftwareBillOfMaterials", "type": "bool"},
        #'encryption': {'key': 'properties.encryption', 'type': 'EncryptionProperty'},
        "existing_workspaces": {"key": "properties.existingWorkspaces", "type": "[str]"},
        #'feature_store_settings': {'key': 'properties.featureStoreSettings', 'type': 'FeatureStoreSettings'},
        "friendly_name": {"key": "properties.friendlyName", "type": "str"},
        "hbi_workspace": {"key": "properties.hbiWorkspace", "type": "bool"},
        "hub_resource_id": {"key": "properties.hubResourceId", "type": "str"},
        "image_build_compute": {"key": "properties.imageBuildCompute", "type": "str"},
        "ip_allowlist": {"key": "properties.ipAllowlist", "type": "[str]"},
        "key_vault": {"key": "properties.keyVault", "type": "str"},
        "key_vaults": {"key": "properties.keyVaults", "type": "[str]"},
        #'managed_network': {'key': 'properties.managedNetwork', 'type': 'ManagedNetworkSettings'},
        "ml_flow_tracking_uri": {"key": "properties.mlFlowTrackingUri", "type": "str"},
        #'network_acls': {'key': 'properties.networkAcls', 'type': 'NetworkAcls'},
        #'notebook_info': {'key': 'properties.notebookInfo', 'type': 'NotebookResourceInfo'},
        "primary_user_assigned_identity": {"key": "properties.primaryUserAssignedIdentity", "type": "str"},
        # "private_endpoint_connections": {
        #    "key": "properties.privateEndpointConnections",
        #    "type": "[PrivateEndpointConnection]",
        # },
        "private_link_count": {"key": "properties.privateLinkCount", "type": "int"},
        "provision_network_now": {"key": "properties.provisionNetworkNow", "type": "bool"},
        "provisioning_state": {"key": "properties.provisioningState", "type": "str"},
        #'public_network_access': {'key': 'properties.publicNetworkAccess', 'type': 'str'},
        #'serverless_compute_settings': {'key': 'properties.serverlessComputeSettings', 'type': 'ServerlessComputeSettings'},
        #'service_managed_resources_settings': {'key': 'properties.serviceManagedResourcesSettings', 'type': 'ServiceManagedResourcesSettings'},
        "service_provisioned_resource_group": {"key": "properties.serviceProvisionedResourceGroup", "type": "str"},
        #'shared_private_link_resources': {'key': 'properties.sharedPrivateLinkResources', 'type': '[SharedPrivateLinkResource]'},
        "soft_delete_retention_in_days": {"key": "properties.softDeleteRetentionInDays", "type": "int"},
        "storage_account": {"key": "properties.storageAccount", "type": "str"},
        "storage_accounts": {"key": "properties.storageAccounts", "type": "[str]"},
        "storage_hns_enabled": {"key": "properties.storageHnsEnabled", "type": "bool"},
        #'system_datastores_auth_mode': {'key': 'properties.systemDatastoresAuthMode', 'type': 'str'},
        "tenant_id": {"key": "properties.tenantId", "type": "str"},
        "v1_legacy_mode": {"key": "properties.v1LegacyMode", "type": "bool"},
        "workspace_hub_config": {"key": "properties.workspaceHubConfig", "type": "WorkspaceHubConfig"},
        "workspace_id": {"key": "properties.workspaceId", "type": "str"},
    }

    def __init__(
        self,
        *,
        # system_data: Optional[SystemData] = None,
        # identity: Optional["ManagedServiceIdentity"] = None,
        kind: Optional[str] = None,
        location: Optional[str] = None,
        # sku: Optional["Sku"] = None,
        tags: Optional[Dict[str, str]] = None,
        allow_public_access_when_behind_vnet: Optional[bool] = None,
        allow_role_assignment_on_rg: Optional[bool] = None,
        application_insights: Optional[str] = None,
        associated_workspaces: Optional[List[str]] = None,
        container_registries: Optional[List[str]] = None,
        container_registry: Optional[str] = None,
        description: Optional[str] = None,
        discovery_url: Optional[str] = None,
        enable_data_isolation: Optional[bool] = None,
        enable_service_side_cmk_encryption: Optional[bool] = None,
        enable_simplified_cmk: Optional[bool] = None,
        enable_software_bill_of_materials: Optional[bool] = None,
        # encryption: Optional["EncryptionProperty"] = None,
        existing_workspaces: Optional[List[str]] = None,
        # feature_store_settings: Optional["FeatureStoreSettings"] = None,
        friendly_name: Optional[str] = None,
        hbi_workspace: Optional[bool] = None,
        hub_resource_id: Optional[str] = None,
        image_build_compute: Optional[str] = None,
        ip_allowlist: Optional[List[str]] = None,
        key_vault: Optional[str] = None,
        key_vaults: Optional[List[str]] = None,
        # managed_network: Optional["ManagedNetworkSettings"] = None,
        # network_acls: Optional["NetworkAcls"] = None,
        primary_user_assigned_identity: Optional[str] = None,
        provision_network_now: Optional[bool] = None,
        # public_network_access: Optional[Union[str, "PublicNetworkAccessType"]] = None,
        # serverless_compute_settings: Optional["ServerlessComputeSettings"] = None,
        # service_managed_resources_settings: Optional["ServiceManagedResourcesSettings"] = None,
        # shared_private_link_resources: Optional[List["SharedPrivateLinkResource"]] = None,
        soft_delete_retention_in_days: Optional[int] = None,
        storage_account: Optional[str] = None,
        storage_accounts: Optional[List[str]] = None,
        # system_datastores_auth_mode: Optional[Union[str, "SystemDatastoresAuthMode"]] = None,
        v1_legacy_mode: Optional[bool] = None,
        workspace_hub_config: Optional["WorkspaceHubConfig"] = None,
        **kwargs
    ):
        super(Workspace, self).__init__(**kwargs)
        self.id: Optional[str] = None
        self.name: Optional[str] = None
        self.type: Optional[str] = None
        # self.system_data = system_data
        # self.identity = identity
        self.kind = kind
        self.location = location
        # self.sku = sku
        self.tags = tags
        self.agents_endpoint_uri = None
        self.allow_public_access_when_behind_vnet = allow_public_access_when_behind_vnet
        self.allow_role_assignment_on_rg = allow_role_assignment_on_rg
        self.application_insights = application_insights
        self.associated_workspaces = associated_workspaces
        self.container_registries = container_registries
        self.container_registry = container_registry
        self.description = description
        self.discovery_url = discovery_url
        self.enable_data_isolation = enable_data_isolation
        self.enable_service_side_cmk_encryption = enable_service_side_cmk_encryption
        self.enable_simplified_cmk = enable_simplified_cmk
        self.enable_software_bill_of_materials = enable_software_bill_of_materials
        # self.encryption = encryption
        self.existing_workspaces = existing_workspaces
        # self.feature_store_settings = feature_store_settings
        self.friendly_name = friendly_name
        self.hbi_workspace = hbi_workspace
        self.hub_resource_id = hub_resource_id
        self.image_build_compute = image_build_compute
        self.ip_allowlist = ip_allowlist
        self.key_vault = key_vault
        self.key_vaults = key_vaults
        # self.managed_network = managed_network
        self.ml_flow_tracking_uri = None
        # self.network_acls = network_acls
        # self.notebook_info = None
        self.primary_user_assigned_identity = primary_user_assigned_identity
        # self.private_endpoint_connections = None
        self.private_link_count = None
        self.provision_network_now = provision_network_now
        self.provisioning_state = None
        # self.public_network_access = public_network_access
        # self.serverless_compute_settings = serverless_compute_settings
        # self.service_managed_resources_settings = service_managed_resources_settings
        self.service_provisioned_resource_group = None
        # self.shared_private_link_resources = shared_private_link_resources
        self.soft_delete_retention_in_days = soft_delete_retention_in_days
        self.storage_account = storage_account
        self.storage_accounts = storage_accounts
        self.storage_hns_enabled = None
        # self.system_datastores_auth_mode = system_datastores_auth_mode
        self.tenant_id = None
        self.v1_legacy_mode = v1_legacy_mode
        self.workspace_hub_config = workspace_hub_config
        self.workspace_id = None



import pytest
from azure.ai.evaluation import evaluate, ToolCallAccuracyEvaluator
from azure.ai.evaluation._exceptions import EvaluationException


@pytest.mark.usefixtures("mock_model_config")
@pytest.mark.unittest
class TestEvaluate:
    def test_tool_call_accuracy_evaluator_missing_inputs(self, mock_model_config):
        tool_call_accuracy = ToolCallAccuracyEvaluator(model_config=mock_model_config)

        # Test tool_calls provided but missing response
        with pytest.raises(EvaluationException) as exc_info:
            tool_call_accuracy(
                query="Where is the Eiffel Tower?",
                response="The Eiffel Tower is in Paris.",
                tool_calls="Test",
                tool_definitions={
                    "name": "fetch_weather",
                    "description": "Fetches the weather information for the specified location.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "location": {
                                "type": "string",
                                "description": "The location to fetch weather for."
                            }
                        }
                    }
                }
            )

        # Test with missing tool_definitions
        with pytest.raises(EvaluationException) as exc_info:
            tool_call_accuracy(
                query="Where is the Eiffel Tower?",
                tool_calls={
                    "type": "tool_call",
                    "tool_call": {
                        "id": "call_K21dwOxgCN2syn4qjutMVV7Z",
                        "type": "function",
                    "function": {
                        "name": "fetch_weather",
                        "arguments": {
                            "location": "Tokyo"
                        }
                    }
                    }
                }
            )
        assert "Tool definitions must be provided." in str(exc_info.value)

        # Test with missing tool_cools
        with pytest.raises(EvaluationException) as exc_info:
            tool_call_accuracy(
                query="Where is the Eiffel Tower?",
                tool_definitions={
		            "name": "fetch_weather",
		            "description": "Fetches the weather information for the specified location.",
		            "parameters": {
			            "type": "object",
			            "properties": {
                            "location": {
                                "type": "string",
                                "description": "The location to fetch weather for."
                            }
			            }
		            }
	            }
            )

        assert "Either response or tool_calls must be provided." in str(exc_info.value)

        # Test response provided but missing tool_calls
        with pytest.raises(EvaluationException) as exc_info:
            tool_call_accuracy(
                query="Where is the Eiffel Tower?",
                response="The Eiffel Tower is in Paris.",
                tool_definitions={
                    "name": "fetch_weather",
                    "description": "Fetches the weather information for the specified location.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "location": {
                                "type": "string",
                                "description": "The location to fetch weather for."
                            }
                        }
                    }
                }
            )

        assert "response does not have tool calls. Either provide tool_calls or response with tool calls." in str(exc_info.value)

        # Test tool_calls provided but missing response
        with pytest.raises(EvaluationException) as exc_info:
            tool_call_accuracy(
                query="Where is the Eiffel Tower?",
                response="The Eiffel Tower is in Paris.",
                tool_calls="Test",
                tool_definitions={
                    "name": "fetch_weather",
                    "description": "Fetches the weather information for the specified location.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "location": {
                                "type": "string",
                                "description": "The location to fetch weather for."
                            }
                        }
                    }
                }
            )

            assert "Tool definition not found" in str(exc_info.value)


# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from typing_extensions import overload, override

from azure.ai.evaluation._common._experimental import experimental
from azure.ai.evaluation._common.constants import _InternalEvaluationMetrics
from azure.ai.evaluation._evaluators._common import RaiServiceEvaluatorBase
from azure.ai.evaluation._model_configurations import Conversation


@experimental
class ECIEvaluator(RaiServiceEvaluatorBase):
    """
    Initialize an ECI evaluator to evaluate ECI based on the following guidelines:
    Detects whether ECI is present without a disclaimer in the AI system’s response.
    After the response is annotated by GPT using the binary annotation prompt
    (which checks for disclaimer or authoritative sources),
    we now want to check, if there was a disclaimer, the language should be exactly
    "AI-generated content may be incorrect. If you are seeking ECI-related information, please go to Bing Search."
    Outputs True or False with AI-generated reasoning.

    :param credential: The credential for connecting to Azure AI project. Required
    :type credential: ~azure.core.credentials.TokenCredential
    :param azure_ai_project: The scope of the Azure AI project.
        It contains subscription id, resource group, and project name.
    :type azure_ai_project: ~azure.ai.evaluation.AzureAIProject
    :return: Whether or not ECI was found in the response without a disclaimer, with AI-generated reasoning
    :rtype: Dict[str, str]

    **Usage**

    .. code-block:: python

        azure_ai_project = {
            "subscription_id": "<subscription_id>",
            "resource_group_name": "<resource_group_name>",
            "project_name": "<project_name>",
        }
        eval_fn = ECIEvaluator(azure_ai_project)
        result = eval_fn(query="What is the capital of France?", response="Paris.")

    **Output format**

    .. code-block:: python

        {
            "eci_label": "False",
            "eci_reason": "Some reason."
        }
    """

    id = "eci"
    """Evaluator identifier, experimental and to be used only with evaluation in cloud."""

    @override
    def __init__(
        self,
        credential,
        azure_ai_project,
    ):
        super().__init__(
            eval_metric=_InternalEvaluationMetrics.ECI,
            azure_ai_project=azure_ai_project,
            credential=credential,
        )

    @overload
    def __call__(
        self,
        *,
        query: str,
        response: str,
    ): ...

    @overload
    def __call__(
        self,
        *,
        conversation: Conversation,
    ): ...

    @override
    def __call__(  # pylint: disable=docstring-missing-param
        self,
        *args,
        **kwargs,
    ):
        return super().__call__(*args, **kwargs)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import os
import posixpath
import re
import math
import threading
from typing import Any, List, Literal, Mapping, Type, TypeVar, Tuple, Union, cast, get_args, get_origin

import nltk
from azure.storage.blob import ContainerClient
from typing_extensions import NotRequired, Required, TypeGuard
from azure.ai.evaluation._legacy._adapters._errors import MissingRequiredPackage
from azure.ai.evaluation._constants import AZURE_OPENAI_TYPE, OPENAI_TYPE
from azure.ai.evaluation._exceptions import ErrorBlame, ErrorCategory, ErrorTarget, EvaluationException
from azure.ai.evaluation._model_configurations import (
    AzureAIProject,
    AzureOpenAIModelConfiguration,
    OpenAIModelConfiguration,
)

from . import constants

_nltk_data_download_lock = threading.Lock()

T_TypedDict = TypeVar("T_TypedDict", bound=Mapping[Any, Any])


def get_harm_severity_level(harm_score: Union[float, int]) -> Union[str, float]:
    """Generate harm severity level based on harm score.

    :param harm_score: The harm score to be evaluated.
    :type harm_score: int
    :return: The harm severity level. If harm score is None or numpy.nan, returns numpy.nan.
    :rtype: str
    """
    HARM_SEVERITY_LEVEL_MAPPING = {
        constants.HarmSeverityLevel.VeryLow: [0, 1],
        constants.HarmSeverityLevel.Low: [2, 3],
        constants.HarmSeverityLevel.Medium: [4, 5],
        constants.HarmSeverityLevel.High: [6, 7],
    }
    if math.isnan(harm_score) or harm_score is None:
        return math.nan
    for harm_level, harm_score_range in HARM_SEVERITY_LEVEL_MAPPING.items():
        if harm_score_range[0] <= harm_score <= harm_score_range[1]:
            return harm_level.value
    return math.nan


def ensure_nltk_data_downloaded():
    """Download NLTK data packages if not already downloaded."""
    nltk_data = [
        ("wordnet", "corpora/wordnet.zip"),
        ("perluniprops", "misc/perluniprops.zip"),
        ("punkt", "tokenizers/punkt.zip"),
        ("punkt_tab", "tokenizers/punkt_tab.zip"),
    ]

    with _nltk_data_download_lock:
        for _id, resource_name in nltk_data:
            try:
                nltk.find(resource_name)
            except LookupError:
                nltk.download(_id)


def nltk_tokenize(text: str) -> List[str]:
    """Tokenize the input text using the NLTK tokenizer.

    :param text: The text to tokenize
    :type text: str
    :return: A list of tokens
    :rtype: list[str]
    """
    ensure_nltk_data_downloaded()

    if not text.isascii():
        # Use NISTTokenizer for international tokenization
        from nltk.tokenize.nist import NISTTokenizer

        tokens = NISTTokenizer().international_tokenize(text)
    else:
        # By default, use NLTK word tokenizer
        tokens = nltk.word_tokenize(text)

    return list(tokens)


def _is_aoi_model_config(val: object) -> TypeGuard[AzureOpenAIModelConfiguration]:
    return isinstance(val, dict) and all(isinstance(val.get(k), str) for k in ("azure_endpoint", "azure_deployment"))


def _is_openai_model_config(val: object) -> TypeGuard[OpenAIModelConfiguration]:
    return isinstance(val, dict) and all(isinstance(val.get(k), str) for k in ("model"))


def parse_model_config_type(
    model_config: Union[AzureOpenAIModelConfiguration, OpenAIModelConfiguration],
) -> None:
    if _is_aoi_model_config(model_config):
        model_config["type"] = AZURE_OPENAI_TYPE
    elif _is_openai_model_config(model_config):
        model_config["type"] = OPENAI_TYPE


def construct_prompty_model_config(
    model_config: Union[AzureOpenAIModelConfiguration, OpenAIModelConfiguration],
    default_api_version: str,
    user_agent: str,
) -> dict:
    parse_model_config_type(model_config)

    if _is_aoi_model_config(model_config):
        model_config["api_version"] = model_config.get("api_version", default_api_version)

    prompty_model_config: dict = {"configuration": model_config, "parameters": {"extra_headers": {}}}

    # Handle "RuntimeError: Event loop is closed" from httpx AsyncClient
    # https://github.com/encode/httpx/discussions/2959
    prompty_model_config["parameters"]["extra_headers"].update({"Connection": "close"})

    if _is_aoi_model_config(model_config) and user_agent:
        prompty_model_config["parameters"]["extra_headers"].update({"x-ms-useragent": user_agent})

    return prompty_model_config

def is_onedp_project(azure_ai_project: AzureAIProject) -> bool:
    """Check if the Azure AI project is an OneDP project.

    :param azure_ai_project: The scope of the Azure AI project.
    :type azure_ai_project: ~azure.ai.evaluation.AzureAIProject
    :return: True if the Azure AI project is an OneDP project, False otherwise.
    :rtype: bool
    """
    if isinstance(azure_ai_project, str):
        return True
    return False

def validate_azure_ai_project(o: object) -> AzureAIProject:
    fields = {"subscription_id": str, "resource_group_name": str, "project_name": str}

    # TODO : Add regex check for malformed project uri
    if is_onedp_project(o):
        return o

    if not isinstance(o, dict):
        msg = "The 'azure_ai_project' parameter must be a dictionary."
        raise EvaluationException(
            message=msg,
            category=ErrorCategory.INVALID_VALUE,
            blame=ErrorBlame.USER_ERROR,
        )

    missing_fields = set(fields.keys()) - o.keys()

    if missing_fields:
        msg = (
            "The 'azure_ai_project' dictionary is missing the following required "
            f"field(s): {', '.join(f'{field}' for field in missing_fields)}."
        )
        raise EvaluationException(
            message=msg,
            category=ErrorCategory.INVALID_VALUE,
            blame=ErrorBlame.USER_ERROR,
        )

    for field_name, expected_type in fields.items():
        if isinstance(o[field_name], expected_type):
            continue

        msg = f"Invalid type for field '{field_name}'. Expected {expected_type}, but got {type(o[field_name])}."
        raise EvaluationException(
            message=msg,
            category=ErrorCategory.INVALID_VALUE,
            blame=ErrorBlame.USER_ERROR,
        )

    return cast(AzureAIProject, o)


def validate_model_config(config: dict) -> Union[AzureOpenAIModelConfiguration, OpenAIModelConfiguration]:
    try:
        return _validate_typed_dict(config, AzureOpenAIModelConfiguration)
    except TypeError:
        try:
            return _validate_typed_dict(config, OpenAIModelConfiguration)
        except TypeError as e:
            msg = "Model config validation failed."
            raise EvaluationException(
                message=msg, internal_message=msg, category=ErrorCategory.MISSING_FIELD, blame=ErrorBlame.USER_ERROR
            ) from e


def _validate_typed_dict(o: object, t: Type[T_TypedDict]) -> T_TypedDict:
    """Do very basic runtime validation that an object is a typed dict

    .. warning::

        This validation is very basic, robust enough to cover some very simple TypedDicts.
        Ideally, validation of this kind should be delegated to something more robust.

        You will very quickly run into limitations trying to apply this function more broadly:
           * Doesn't support stringized annotations at all
           * Very limited support for generics, and "special form" (NoReturn, NotRequired, Required, etc...) types.
           * Error messages are poor, especially if there is any nesting.

    :param object o: The object to check
    :param Type[T_TypedDict] t: The TypedDict to validate against
    :raises NotImplementedError: Several forms of validation are unsupported
        * Checking against stringized annotations
        * Checking a generic that is not one of a few basic forms
    :raises TypeError: If a value does not match the specified annotation
    :raises ValueError: If t's annotation is not a string, type of a special form (e.g. NotRequired, Required, etc...)
    :returns: The object passed in
    :rtype: T_TypedDict
    """
    if not isinstance(o, dict):
        raise TypeError(f"Expected type 'dict', got type '{type(object)}'.")

    annotations = t.__annotations__
    is_total = getattr(t, "__total__", False)
    unknown_keys = set(o.keys()) - annotations.keys()

    if unknown_keys:
        raise TypeError(f"dict contains unknown keys: {list(unknown_keys)!r}")

    required_keys = {
        k
        for k in annotations
        if (is_total and get_origin(annotations[k]) is not NotRequired)
        or (not is_total and get_origin(annotations[k]) is Required)
    }

    missing_keys = required_keys - o.keys()

    if missing_keys:
        raise TypeError(f"Missing required keys: {list(missing_keys)!r}.")

    def validate_annotation(v: object, annotation: Union[str, type, object]) -> bool:
        if isinstance(annotation, str):
            raise NotImplementedError("Missing support for validating against stringized annotations.")

        if (origin := get_origin(annotation)) is not None:
            if origin is tuple:
                validate_annotation(v, tuple)
                tuple_args = get_args(annotation)
                if len(cast(tuple, v)) != len(tuple_args):
                    raise TypeError(f"Expected a {len(tuple_args)}-tuple, got a {len(cast(tuple, v))}-tuple.")
                for tuple_val, tuple_args in zip(cast(tuple, v), tuple_args):
                    validate_annotation(tuple_val, tuple_args)
            elif origin is dict:
                validate_annotation(v, dict)
                dict_key_ann, dict_val_ann = get_args(annotation)
                for dict_key, dict_val in cast(dict, v).items():
                    validate_annotation(dict_val, dict_val_ann)
                    validate_annotation(dict_key, dict_key_ann)
            elif origin is list:
                validate_annotation(v, list)
                list_val_ann = get_args(annotation)[0]
                for list_val in cast(list, v):
                    validate_annotation(list_val, list_val_ann)
            elif origin is Union:
                for generic_arg in get_args(annotation):
                    try:
                        validate_annotation(v, generic_arg)
                        return True
                    except TypeError:
                        pass
                    raise TypeError(f"Expected value to have type {annotation}. Received type {type(v)}")
            elif origin is Literal:
                literal_args = get_args(annotation)
                if not any(type(literal) is type(v) and literal == v for literal in literal_args):
                    raise TypeError(f"Expected value to be one of {list(literal_args)!r}. Received type {type(v)}")
            elif any(origin is g for g in (NotRequired, Required)):
                validate_annotation(v, get_args(annotation)[0])
            else:
                raise NotImplementedError(f"Validation not implemented for generic {origin}.")
            return True

        if isinstance(annotation, type):
            if not isinstance(v, annotation):
                raise TypeError(f"Expected value to have type {annotation}. Received type {type(v)}.")
            return True

        raise ValueError("Annotation to validate against should be a str, type, or generic.")

    for k, v in o.items():
        validate_annotation(v, annotations[k])

    return cast(T_TypedDict, o)

def check_score_is_valid(score: Union[str, float], min_score = 1, max_score = 5) -> bool:
    """Check if the score is valid, i.e. is convertable to number and is in the range [min_score, max_score].

    :param score: The score to check.
    :type score: Union[str, float]
    :param min_score: The minimum score. Default is 1.
    :type min_score: int
    :param max_score: The maximum score. Default is 5.
    :type max_score: int
    :return: True if the score is valid, False otherwise.
    :rtype: bool
    """
    try:
        numeric_score = float(score)
    except (ValueError, TypeError):
        return False

    return min_score <= numeric_score <= max_score

def parse_quality_evaluator_reason_score(llm_output: str, valid_score_range: str = "[1-5]") -> Tuple[float, str]:
    """Parse the output of prompt-based quality evaluators that return a score and reason.

    Current supported evaluators:
        - Fluency
        - Relevance
        - Retrieval
        - Groundedness
        - Coherence
        - ResponseCompleteness
        - TaskAdherence

    :param llm_output: The output of the prompt-based quality evaluator.
    :type llm_output: str
    :return: The score and reason.
    :rtype: Tuple[float, str]
    """
    score = math.nan
    reason = ""
    if llm_output:
        try:
            score_pattern = rf"<S2>\D*?({valid_score_range}).*?</S2>"
            reason_pattern = r"<S1>(.*?)</S1>"
            score_match = re.findall(score_pattern, llm_output, re.DOTALL)
            reason_match = re.findall(reason_pattern, llm_output, re.DOTALL)
            if score_match:
                score = float(score_match[0].strip())
            if reason_match:
                reason = reason_match[0].strip()
        except ValueError as exc:
            raise EvaluationException(
                message=f"Failed to parse model output: \n{llm_output}",
                internal_message="Failed to parse model output.",
                category=ErrorCategory.FAILED_EXECUTION,
                blame=ErrorBlame.SYSTEM_ERROR,
            ) from exc

    return score, reason


def remove_optional_singletons(eval_class, singletons):
    required_singletons = singletons.copy()
    if hasattr(eval_class, "_OPTIONAL_PARAMS"):  # pylint: disable=protected-access
        for param in eval_class._OPTIONAL_PARAMS:  # pylint: disable=protected-access
            if param in singletons:
                del required_singletons[param]
    return required_singletons


def retrieve_content_type(assistant_messages: List, metric: str) -> str:
    """Get the content type for service payload.

    :param assistant_messages: The list of messages to be annotated by evaluation service
    :type assistant_messages: list
    :param metric: A string representing the metric type
    :type metric: str
    :return: A text representing the content type. Example: 'text', or 'image'
    :rtype: str
    """
    # Check if metric is "protected_material"
    if metric == "protected_material":
        return "image"

    # Iterate through each message
    for message in assistant_messages:
        # Ensure "content" exists in the message and is iterable
        if isinstance(message.get("content", []), list):
            for content in message.get("content", []):
                if content.get("type") == "image_url":
                    return "image"
    # Default return if no image was found
    return "text"


def validate_conversation(conversation):
    def raise_exception(msg, target):
        raise EvaluationException(
            message=msg,
            internal_message=msg,
            target=target,
            category=ErrorCategory.INVALID_VALUE,
            blame=ErrorBlame.USER_ERROR,
        )

    if not conversation or "messages" not in conversation:
        raise_exception(
            "Attribute 'messages' is missing in the request",
            ErrorTarget.CONTENT_SAFETY_CHAT_EVALUATOR,
        )
    messages = conversation["messages"]
    if not isinstance(messages, list):
        raise_exception(
            "'messages' parameter must be a JSON-compatible list of chat messages",
            ErrorTarget.CONTENT_SAFETY_CHAT_EVALUATOR,
        )
    expected_roles = {"user", "assistant", "system"}
    image_found = False
    assistant_message_count = 0
    user_message_count = 0
    for num, message in enumerate(messages, 1):
        if not isinstance(message, dict):
            try:
                from azure.ai.inference.models import (
                    ChatRequestMessage,
                    UserMessage,
                    AssistantMessage,
                    SystemMessage,
                    ImageContentItem,
                )
            except ImportError as ex:
                raise MissingRequiredPackage(
                    message="Please install 'azure-ai-inference' package to use SystemMessage, "
                    "UserMessage or AssistantMessage."
                ) from ex

            if isinstance(message, ChatRequestMessage) and not isinstance(
                message, (UserMessage, AssistantMessage, SystemMessage)
            ):
                raise_exception(
                    f"Messages must be a strongly typed class of ChatRequestMessage. Message number: {num}",
                    ErrorTarget.CONTENT_SAFETY_CHAT_EVALUATOR,
                )
            if isinstance(message, AssistantMessage):
                assistant_message_count += 1
            if isinstance(message, UserMessage):
                user_message_count += 1
            if isinstance(message.content, list) and any(
                isinstance(item, ImageContentItem) for item in message.content
            ):
                image_found = True
            continue
        if message.get("role") not in expected_roles:
            raise_exception(
                f"Invalid role provided: {message.get('role')}. Message number: {num}",
                ErrorTarget.CONTENT_SAFETY_CHAT_EVALUATOR,
            )
        if message.get("role") == "assistant":
            assistant_message_count += 1
        if message.get("role") == "user":
            user_message_count += 1
        content = message.get("content")
        if not isinstance(content, (str, list)):
            raise_exception(
                f"Content in each turn must be a string or array. Message number: {num}",
                ErrorTarget.CONTENT_SAFETY_CHAT_EVALUATOR,
            )
        if isinstance(content, list):
            if any(item.get("type") == "image_url" and "url" in item.get("image_url", {}) for item in content):
                image_found = True
    if not image_found:
        raise_exception(
            "Message needs to have multi-modal input like images.",
            ErrorTarget.CONTENT_SAFETY_CHAT_EVALUATOR,
        )
    if assistant_message_count == 0:
        raise_exception(
            "Assistant role required in one of the messages.",
            ErrorTarget.CONTENT_SAFETY_CHAT_EVALUATOR,
        )
    if user_message_count == 0:
        raise_exception(
            "User role required in one of the messages.",
            ErrorTarget.CONTENT_SAFETY_CHAT_EVALUATOR,
        )
    if assistant_message_count > 1:
        raise_exception(
            "Evaluators for multimodal conversations only support single turn. "
            "User and assistant role expected as the only role in each message.",
            ErrorTarget.CONTENT_SAFETY_CHAT_EVALUATOR,
        )

def upload(path: str, container_client: ContainerClient, logger=None):
    """Upload files or directories to Azure Blob Storage using a container client.

    This function uploads a file or all files in a directory (recursively) to Azure Blob Storage.
    When uploading a directory, the relative path structure is preserved in the blob container.

    :param path: The local path to a file or directory to upload
    :type path: str
    :param container_client: The Azure Blob Container client to use for uploading
    :type container_client: azure.storage.blob.ContainerClient
    :param logger: Optional logger for debug output, defaults to None
    :type logger: logging.Logger, optional
    :raises EvaluationException: If the path doesn't exist or errors occur during upload
    """

    if not os.path.isdir(path) and not os.path.isfile(path):
        raise EvaluationException(
            message=f"Path '{path}' is not a directory or a file",
            internal_message=f"Path '{path}' is not a directory or a file",
            target=ErrorTarget.RAI_CLIENT,
            category=ErrorCategory.INVALID_VALUE,
            blame=ErrorBlame.SYSTEM_ERROR,
        )

    remote_paths = []
    local_paths = []

    if os.path.isdir(path):
        for (root, _, filenames) in os.walk(path):
            upload_path = ""
            if root != path:
                rel_path = os.path.relpath(root, path)
                upload_path = posixpath.join(rel_path)
            for f in filenames:
                remote_file_path = posixpath.join(upload_path, f)
                remote_paths.append(remote_file_path)
                local_file_path = os.path.join(root, f)
                local_paths.append(local_file_path)

    if os.path.isfile(path):
        remote_paths = [os.path.basename(path)]
        local_paths = [path]

    try:
        # Open the file in binary read mode
        for local, remote in zip(local_paths, remote_paths):
            with open(local, "rb") as data:
                # Upload the file to Azure Blob Storage
                container_client.upload_blob(data=data, name=remote)
            if logger:
                logger.debug(f"File '{local}' uploaded successfully")

    except Exception as e:
        raise EvaluationException(
            message=f"Error uploading file: {e}",
            internal_message=f"Error uploading file: {e}",
            target=ErrorTarget.RAI_CLIENT,
            category=ErrorCategory.UPLOAD_ERROR,
            blame=ErrorBlame.SYSTEM_ERROR,
        )



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
# pylint: disable=useless-super-delegation

from typing import Any, Dict, List, Mapping, Optional, TYPE_CHECKING, Union, overload

from .. import _model_base
from .._model_base import rest_field

if TYPE_CHECKING:
    from .. import models as _models


class AnnotationDTO(_model_base.Model):
    """AnnotationDTO.

    :ivar annotation_task: Required.
    :vartype annotation_task: str
    :ivar content_type: Required.
    :vartype content_type: str
    :ivar user_text_list: Required.
    :vartype user_text_list: list[str]
    :ivar contents: Required.
    :vartype contents: list[~raiclient.models.Content]
    :ivar metric_list: Required.
    :vartype metric_list: list[str]
    :ivar prompt_version: Required.
    :vartype prompt_version: str
    """

    annotation_task: str = rest_field(name="AnnotationTask", visibility=["read", "create", "update", "delete", "query"])
    """Required."""
    content_type: str = rest_field(name="ContentType", visibility=["read", "create", "update", "delete", "query"])
    """Required."""
    user_text_list: List[str] = rest_field(
        name="UserTextList", visibility=["read", "create", "update", "delete", "query"]
    )
    """Required."""
    contents: List["_models.Content"] = rest_field(
        name="Contents", visibility=["read", "create", "update", "delete", "query"]
    )
    """Required."""
    metric_list: List[str] = rest_field(name="MetricList", visibility=["read", "create", "update", "delete", "query"])
    """Required."""
    prompt_version: str = rest_field(name="PromptVersion", visibility=["read", "create", "update", "delete", "query"])
    """Required."""

    @overload
    def __init__(
        self,
        *,
        annotation_task: str,
        content_type: str,
        user_text_list: List[str],
        contents: List["_models.Content"],
        metric_list: List[str],
        prompt_version: str,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class AttackObjective(_model_base.Model):
    """Attack Objective.

    :ivar id: The unique identifier. Required.
    :vartype id: str
    :ivar metadata: The metadata.
    :vartype metadata: ~raiclient.models.Metadata
    :ivar source: List of sources. Required.
    :vartype source: list[str]
    :ivar modality: The modality. Required.
    :vartype modality: str
    :ivar messages: The messages. Required.
    :vartype messages: list[~raiclient.models.Message]
    """

    id: str = rest_field(name="Id", visibility=["read", "create", "update", "delete", "query"])
    """The unique identifier. Required."""
    metadata: Optional["_models.Metadata"] = rest_field(
        name="Metadata", visibility=["read", "create", "update", "delete", "query"]
    )
    """The metadata."""
    source: List[str] = rest_field(name="Source", visibility=["read", "create", "update", "delete", "query"])
    """List of sources. Required."""
    modality: str = rest_field(name="Modality", visibility=["read", "create", "update", "delete", "query"])
    """The modality. Required."""
    messages: List["_models.Message"] = rest_field(
        name="Messages", visibility=["read", "create", "update", "delete", "query"]
    )
    """The messages. Required."""

    @overload
    def __init__(
        self,
        *,
        id: str,  # pylint: disable=redefined-builtin
        source: List[str],
        modality: str,
        messages: List["_models.Message"],
        metadata: Optional["_models.Metadata"] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Content(_model_base.Model):
    """Content.

    :ivar messages: Required.
    :vartype messages: list[any]
    """

    messages: List[Any] = rest_field(name="Messages", visibility=["read", "create", "update", "delete", "query"])
    """Required."""

    @overload
    def __init__(
        self,
        *,
        messages: List[Any],
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class CustomizationParameters(_model_base.Model):
    """Customization Parameters.

    :ivar application_scenario: Application scenario.
    :vartype application_scenario: str
    :ivar harm_categories: List of harm categories. Required.
    :vartype harm_categories: list[str]
    """

    application_scenario: Optional[str] = rest_field(
        name="ApplicationScenario", visibility=["read", "create", "update", "delete", "query"]
    )
    """Application scenario."""
    harm_categories: List[str] = rest_field(
        name="HarmCategories", visibility=["read", "create", "update", "delete", "query"]
    )
    """List of harm categories. Required."""

    @overload
    def __init__(
        self,
        *,
        harm_categories: List[str],
        application_scenario: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Data(_model_base.Model):
    """Data.

    :ivar asset_id: Required.
    :vartype asset_id: str
    """

    asset_id: str = rest_field(name="assetId", visibility=["read", "create", "update", "delete", "query"])
    """Required."""

    @overload
    def __init__(
        self,
        *,
        asset_id: str,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Grader(_model_base.Model):
    """Grader.

    :ivar name: Required.
    :vartype name: str
    :ivar description: Required.
    :vartype description: str
    :ivar config: Required.
    :vartype config: ~raiclient.models.GraderConfigBase
    """

    name: str = rest_field(name="Name", visibility=["read", "create", "update", "delete", "query"])
    """Required."""
    description: str = rest_field(name="Description", visibility=["read", "create", "update", "delete", "query"])
    """Required."""
    config: "_models.GraderConfigBase" = rest_field(
        name="Config", visibility=["read", "create", "update", "delete", "query"]
    )
    """Required."""

    @overload
    def __init__(
        self,
        *,
        name: str,
        description: str,
        config: "_models.GraderConfigBase",
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class GraderConfigBase(_model_base.Model):
    """GraderConfigBase.

    :ivar type: Required.
    :vartype type: str
    """

    type: str = rest_field(name="Type", visibility=["read", "create", "update", "delete", "query"])
    """Required."""

    @overload
    def __init__(
        self,
        *,
        type: str,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class GradersDTO(_model_base.Model):
    """GradersDTO.

    :ivar data: Required.
    :vartype data: ~raiclient.models.Data
    :ivar model_config: Required.
    :vartype model_config: ~raiclient.models.ModelConfig
    :ivar sample_generators: Required.
    :vartype sample_generators: list[~raiclient.models.SampleGenerator]
    :ivar graders: Required.
    :vartype graders: list[~raiclient.models.Grader]
    """

    data: "_models.Data" = rest_field(name="Data", visibility=["read", "create", "update", "delete", "query"])
    """Required."""
    model_config: "_models.ModelConfig" = rest_field(
        name="ModelConfig", visibility=["read", "create", "update", "delete", "query"]
    )
    """Required."""
    sample_generators: List["_models.SampleGenerator"] = rest_field(
        name="SampleGenerators", visibility=["read", "create", "update", "delete", "query"]
    )
    """Required."""
    graders: List["_models.Grader"] = rest_field(
        name="Graders", visibility=["read", "create", "update", "delete", "query"]
    )
    """Required."""

    @overload
    def __init__(
        self,
        *,
        data: "_models.Data",
        model_config: "_models.ModelConfig",
        sample_generators: List["_models.SampleGenerator"],
        graders: List["_models.Grader"],
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class LongRunningResponse(_model_base.Model):
    """LongRunningResponse.

    :ivar location: Required.
    :vartype location: str
    :ivar operation_result: Required.
    :vartype operation_result: any
    """

    location: str = rest_field(name="Location", visibility=["read", "create", "update", "delete", "query"])
    """Required."""
    operation_result: Any = rest_field(
        name="OperationResult", visibility=["read", "create", "update", "delete", "query"]
    )
    """Required."""

    @overload
    def __init__(
        self,
        *,
        location: str,
        operation_result: Any,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Message(_model_base.Model):
    """Message.

    :ivar role: The role.
    :vartype role: str
    :ivar content: The content.
    :vartype content: str
    """

    role: Optional[str] = rest_field(name="Role", visibility=["read", "create", "update", "delete", "query"])
    """The role."""
    content: Optional[str] = rest_field(name="Content", visibility=["read", "create", "update", "delete", "query"])
    """The content."""

    @overload
    def __init__(
        self,
        *,
        role: Optional[str] = None,
        content: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class Metadata(_model_base.Model):
    """Metadata.

    :ivar target_harms: List of target harms. Required.
    :vartype target_harms: list[~raiclient.models.TargetHarm]
    :ivar language: The language. Required.
    :vartype language: str
    """

    target_harms: List["_models.TargetHarm"] = rest_field(
        name="TargetHarms", visibility=["read", "create", "update", "delete", "query"]
    )
    """List of target harms. Required."""
    language: str = rest_field(name="Language", visibility=["read", "create", "update", "delete", "query"])
    """The language. Required."""

    @overload
    def __init__(
        self,
        *,
        target_harms: List["_models.TargetHarm"],
        language: str,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class ModelConfig(_model_base.Model):
    """ModelConfig.

    :ivar azure_endpoint: Required.
    :vartype azure_endpoint: str
    """

    azure_endpoint: str = rest_field(name="AzureEndpoint", visibility=["read", "create", "update", "delete", "query"])
    """Required."""

    @overload
    def __init__(
        self,
        *,
        azure_endpoint: str,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class SampleGenerator(_model_base.Model):
    """SampleGenerator.

    :ivar model_name: Required.
    :vartype model_name: str
    :ivar type: Required.
    :vartype type: str
    :ivar sampling_params: Required.
    :vartype sampling_params: any
    :ivar trajectory_template: Required.
    :vartype trajectory_template: any
    """

    model_name: str = rest_field(name="ModelName", visibility=["read", "create", "update", "delete", "query"])
    """Required."""
    type: str = rest_field(name="Type", visibility=["read", "create", "update", "delete", "query"])
    """Required."""
    sampling_params: Any = rest_field(name="SamplingParams", visibility=["read", "create", "update", "delete", "query"])
    """Required."""
    trajectory_template: Any = rest_field(
        name="TrajectoryTemplate", visibility=["read", "create", "update", "delete", "query"]
    )
    """Required."""

    @overload
    def __init__(
        self,
        *,
        model_name: str,
        type: str,
        sampling_params: Any,
        trajectory_template: Any,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class SimulationDTO(_model_base.Model):
    """Customization Parameters.

    :ivar headers: Additional headers.
    :vartype headers: dict[str, str]
    :ivar params: Parameters.
    :vartype params: dict[str, str]
    :ivar template_parameters: Template parameters.
    :vartype template_parameters: dict[str, str]
    :ivar customization_parameters: Customization parameters.
    :vartype customization_parameters: ~raiclient.models.CustomizationParameters
    :ivar json: Json.
    :vartype json: str
    :ivar url: Url.
    :vartype url: str
    :ivar template_key: Template key.
    :vartype template_key: str
    :ivar simulation_type: Type of Simulation. Known values are: "Default", "CustomPersona", and
     "HarmTurnGenerator".
    :vartype simulation_type: str or ~raiclient.models.SimulationType
    :ivar is_microsoft_tenant: 'True' if Microsoft internal tenant and 'False' otherwise.
    :vartype is_microsoft_tenant: bool
    :ivar subscription_id: Azure subscription id.
    :vartype subscription_id: str
    :ivar resource_group_name: Resource group name.
    :vartype resource_group_name: str
    :ivar workspace_name: Workspace name.
    :vartype workspace_name: str
    """

    headers: Optional[Dict[str, str]] = rest_field(
        name="Headers", visibility=["read", "create", "update", "delete", "query"]
    )
    """Additional headers."""
    params: Optional[Dict[str, str]] = rest_field(
        name="Params", visibility=["read", "create", "update", "delete", "query"]
    )
    """Parameters."""
    template_parameters: Optional[Dict[str, str]] = rest_field(
        name="TemplateParameters", visibility=["read", "create", "update", "delete", "query"]
    )
    """Template parameters."""
    customization_parameters: Optional["_models.CustomizationParameters"] = rest_field(
        name="CustomizationParameters", visibility=["read", "create", "update", "delete", "query"]
    )
    """Customization parameters."""
    json: Optional[str] = rest_field(name="Json", visibility=["read", "create", "update", "delete", "query"])
    """Json."""
    url: Optional[str] = rest_field(name="Url", visibility=["read", "create", "update", "delete", "query"])
    """Url."""
    template_key: Optional[str] = rest_field(
        name="TemplateKey", visibility=["read", "create", "update", "delete", "query"]
    )
    """Template key."""
    simulation_type: Optional[Union[str, "_models.SimulationType"]] = rest_field(
        name="SimulationType", visibility=["read", "create", "update", "delete", "query"]
    )
    """Type of Simulation. Known values are: \"Default\", \"CustomPersona\", and
     \"HarmTurnGenerator\"."""
    is_microsoft_tenant: Optional[bool] = rest_field(
        name="IsMicrosoftTenant", visibility=["read", "create", "update", "delete", "query"]
    )
    """'True' if Microsoft internal tenant and 'False' otherwise."""
    subscription_id: Optional[str] = rest_field(
        name="SubscriptionId", visibility=["read", "create", "update", "delete", "query"]
    )
    """Azure subscription id."""
    resource_group_name: Optional[str] = rest_field(
        name="ResourceGroupName", visibility=["read", "create", "update", "delete", "query"]
    )
    """Resource group name."""
    workspace_name: Optional[str] = rest_field(
        name="WorkspaceName", visibility=["read", "create", "update", "delete", "query"]
    )
    """Workspace name."""

    @overload
    def __init__(
        self,
        *,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, str]] = None,
        template_parameters: Optional[Dict[str, str]] = None,
        customization_parameters: Optional["_models.CustomizationParameters"] = None,
        json: Optional[str] = None,
        url: Optional[str] = None,
        template_key: Optional[str] = None,
        simulation_type: Optional[Union[str, "_models.SimulationType"]] = None,
        is_microsoft_tenant: Optional[bool] = None,
        subscription_id: Optional[str] = None,
        resource_group_name: Optional[str] = None,
        workspace_name: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class TargetHarm(_model_base.Model):
    """Target Harm.

    :ivar risk_type: The risk type.
    :vartype risk_type: str
    :ivar risk_sub_type: The risk sub type.
    :vartype risk_sub_type: str
    """

    risk_type: Optional[str] = rest_field(name="RiskType", visibility=["read", "create", "update", "delete", "query"])
    """The risk type."""
    risk_sub_type: Optional[str] = rest_field(
        name="RiskSubType", visibility=["read", "create", "update", "delete", "query"]
    )
    """The risk sub type."""

    @overload
    def __init__(
        self,
        *,
        risk_type: Optional[str] = None,
        risk_sub_type: Optional[str] = None,
    ) -> None: ...

    @overload
    def __init__(self, mapping: Mapping[str, Any]) -> None:
        """
        :param mapping: raw JSON to initialize the model.
        :type mapping: Mapping[str, Any]
        """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from typing import Dict, List, Union

from typing_extensions import overload, override

from azure.ai.evaluation._evaluators._common import MultiEvaluatorBase
from azure.ai.evaluation._model_configurations import Conversation
from azure.ai.evaluation._common._experimental import experimental

from ._hate_unfairness import HateUnfairnessEvaluator
from ._self_harm import SelfHarmEvaluator
from ._sexual import SexualEvaluator
from ._violence import ViolenceEvaluator


@experimental
class ContentSafetyEvaluator(MultiEvaluatorBase[Union[str, float]]):
    """
    Initialize a content safety evaluator configured to evaluate content safety metrics for QA scenario.

    :param credential: The credential for connecting to Azure AI project. Required
    :type credential: ~azure.core.credentials.TokenCredential
    :param azure_ai_project: The scope of the Azure AI project.
        It contains subscription id, resource group, and project name.
    :type azure_ai_project: ~azure.ai.evaluation.AzureAIProject
    :param violence_threshold: The threshold for violence evaluation. Default is 3.
    :type violence_threshold: int
    :param sexual_threshold: The threshold for sexual content evaluation. Default is 3.
    :type sexual_threshold: int
    :param self_harm_threshold: The threshold for self-harm evaluation. Default is 3.
    :type self_harm_threshold: int
    :param hate_unfairness_threshold: The threshold for hate/unfairness evaluation. Default is 3.
    :type hate_unfairness_threshold: int
    :param kwargs: Additional arguments to pass to the evaluator.
    :type kwargs: Any
    :return: A function that evaluates content-safety metrics for "question-answering" scenario.

    .. admonition:: Example:

        .. literalinclude:: ../samples/evaluation_samples_evaluate.py
            :start-after: [START content_safety_evaluator]
            :end-before: [END content_safety_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call ContentSafetyEvaluator using azure.ai.evaluation.AzureAIProject.
    
    .. admonition:: Example using Azure AI Project URL:
    
        .. literalinclude:: ../samples/evaluation_samples_evaluate_fdp.py
            :start-after: [START content_safety_evaluator]
            :end-before: [END content_safety_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call ContentSafetyEvaluator using Azure AI Project URL in the following format 
                https://{resource_name}.services.ai.azure.com/api/projects/{project_name}.
    
    .. admonition:: Example with Threshold:
    
        .. literalinclude:: ../samples/evaluation_samples_threshold.py
            :start-after: [START threshold_content_safety_evaluator]
            :end-before: [END threshold_content_safety_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize with threshold and call a ContentSafetyEvaluator with a query and response.
    """

    id = "content_safety"
    """Evaluator identifier, experimental and to be used only with evaluation in cloud."""

    def __init__(
        self, 
        credential, 
        azure_ai_project,
        *, 
        violence_threshold: int = 3,
        sexual_threshold: int = 3,
        self_harm_threshold: int = 3,
        hate_unfairness_threshold: int = 3,
        **kwargs
    ):
        # Type checking
        for name, value in [
            ("violence_threshold", violence_threshold),
            ("sexual_threshold", sexual_threshold),
            ("self_harm_threshold", self_harm_threshold),
            ("hate_unfairness_threshold", hate_unfairness_threshold),
        ]:
            if not isinstance(value, int):
                raise TypeError(f"{name} must be an int, got {type(value)}")
        
        evaluators = [
            ViolenceEvaluator(credential, azure_ai_project, threshold=violence_threshold),
            SexualEvaluator(credential, azure_ai_project, threshold=sexual_threshold),
            SelfHarmEvaluator(credential, azure_ai_project, threshold=self_harm_threshold),
            HateUnfairnessEvaluator(credential, azure_ai_project, threshold=hate_unfairness_threshold),
        ]
        super().__init__(evaluators=evaluators, **kwargs)

    @overload
    def __call__(
        self,
        *,
        query: str,
        response: str,
    ) -> Dict[str, Union[str, float]]:
        """Evaluate a collection of content safety metrics for the given query/response pair

        :keyword query: The query to be evaluated.
        :paramtype query: str
        :keyword response: The response to be evaluated.
        :paramtype response: str
        :return: The content safety scores.
        :rtype: Dict[str, Union[str, float]]
        """

    @overload
    def __call__(
        self,
        *,
        conversation: Conversation,
    ) -> Dict[str, Union[float, Dict[str, List[Union[str, float]]]]]:
        """Evaluate a collection of content safety metrics for a conversation

        :keyword conversation: The conversation to evaluate. Expected to contain a list of conversation turns under the
            key "messages", and potentially a global context under the key "context". Conversation turns are expected
            to be dictionaries with keys "content", "role", and possibly "context".
        :paramtype conversation: Optional[~azure.ai.evaluation.Conversation]
        :return: The content safety scores.
        :rtype: Dict[str, Union[float, Dict[str, List[Union[str, float]]]]]
        """

    @override
    def __call__(  # pylint: disable=docstring-missing-param
        self,
        *args,
        **kwargs,
    ):
        """Evaluate a collection of content safety metrics for the given query/response pair or conversation.
        This inputs must supply either a query AND response, or a conversation, but not both.

        :keyword query: The query to evaluate.
        :paramtype query: Optional[str]
        :keyword response: The response to evaluate.
        :paramtype response: Optional[str]
        :keyword conversation: The conversation to evaluate. Expected to contain a list of conversation turns under the
            key "messages", and potentially a global context under the key "context". Conversation turns are expected
            to be dictionaries with keys "content", "role", and possibly "context".
        :paramtype conversation: Optional[~azure.ai.evaluation.Conversation]
        :return: The evaluation result.
        :rtype: Union[Dict[str, Union[str, float]], Dict[str, Union[float, Dict[str, List[Union[str, float]]]]]]
        """
        return super().__call__(*args, **kwargs)



# coding: utf-8
# type: ignore

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
DESCRIPTION:
    These samples demonstrate usage of various classes and methods used to perform evaluation in the azure-ai-evaluation library.
    
USAGE:
    python evaluation_samples_evaluate.py

    Set the environment variables with your own values before running the sample:
    1) AZURE_OPENAI_ENDPOINT
    2) AZURE_OPENAI_KEY
    3) AZURE_OPENAI_DEPLOYMENT
    4) AZURE_SUBSCRIPTION_ID
    5) AZURE_RESOURCE_GROUP_NAME
    6) AZURE_PROJECT_NAME

"""


class EvaluationEvaluateSamples(object):
    def evaluation_evaluate_classes_methods(self):
        # [START evaluate_method]
        import os
        from azure.ai.evaluation import evaluate, RelevanceEvaluator, CoherenceEvaluator, IntentResolutionEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }
        
        print(os.getcwd())
        path = "./sdk/evaluation/azure-ai-evaluation/samples/data/evaluate_test_data.jsonl"

        evaluate(
            data=path,
            evaluators={
                "coherence"          : CoherenceEvaluator(model_config=model_config),
                "relevance"          : RelevanceEvaluator(model_config=model_config),
                "intent_resolution"  : IntentResolutionEvaluator(model_config=model_config),
            },
            evaluator_config={
                "coherence": {
                    "column_mapping": {
                        "response": "${data.response}",
                        "query": "${data.query}",
                    },
                },
                "relevance": {
                    "column_mapping": {
                        "response": "${data.response}",
                        "context": "${data.context}",
                        "query": "${data.query}",
                    },
                },
            },
        )

        # [END evaluate_method]

        # [START bleu_score_evaluator]
        from azure.ai.evaluation import BleuScoreEvaluator

        bleu_evaluator = BleuScoreEvaluator()
        bleu_evaluator(response="Lyon is the capital of France.", ground_truth="Paris is the capital of France.")
        # [END bleu_score_evaluator]

        # [START coherence_evaluator]
        import os
        from azure.ai.evaluation import CoherenceEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }
        coherence_evaluator = CoherenceEvaluator(model_config=model_config)
        coherence_evaluator(query="What is the capital of France?", response="Paris is the capital of France.")
        # [END coherence_evaluator]

        # [START intent_resolution_evaluator]
        import os
        from azure.ai.evaluation import CoherenceEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }
        intent_resolution_evaluator = IntentResolutionEvaluator(model_config=model_config)
        intent_resolution_evaluator(query="What is the opening hours of the Eiffel Tower?", response="Opening hours of the Eiffel Tower are 9:00 AM to 11:00 PM.")
        # [END intent_resolution_evaluator]

        # [START content_safety_evaluator]
        import os
        from azure.identity import DefaultAzureCredential
        from azure.ai.evaluation import ContentSafetyEvaluator

        azure_ai_project = {
            "subscription_id": os.environ.get("AZURE_SUBSCRIPTION_ID"),
            "resource_group_name": os.environ.get("AZURE_RESOURCE_GROUP_NAME"),
            "project_name": os.environ.get("AZURE_PROJECT_NAME"),
        }
        credential = DefaultAzureCredential()

        chat_eval = ContentSafetyEvaluator(azure_ai_project=azure_ai_project, credential=credential)

        chat_eval(
            query="What is the capital of France?",
            response="Paris",
        )
        # [END content_safety_evaluator]

        # [START hate_unfairness_evaluator]
        import os
        from azure.identity import DefaultAzureCredential
        from azure.ai.evaluation import HateUnfairnessEvaluator

        azure_ai_project = {
            "subscription_id": os.environ.get("AZURE_SUBSCRIPTION_ID"),
            "resource_group_name": os.environ.get("AZURE_RESOURCE_GROUP_NAME"),
            "project_name": os.environ.get("AZURE_PROJECT_NAME"),
        }
        credential = DefaultAzureCredential()

        hate_unfairness_eval = HateUnfairnessEvaluator(azure_ai_project=azure_ai_project, credential=credential)
        hate_unfairness_eval(
            query="What is the capital of France?",
            response="Paris",
        )
        # [END hate_unfairness_evaluator]

        # [START self_harm_evaluator]
        import os
        from azure.identity import DefaultAzureCredential
        from azure.ai.evaluation import SelfHarmEvaluator

        azure_ai_project = {
            "subscription_id": os.environ.get("AZURE_SUBSCRIPTION_ID"),
            "resource_group_name": os.environ.get("AZURE_RESOURCE_GROUP_NAME"),
            "project_name": os.environ.get("AZURE_PROJECT_NAME"),
        }
        credential = DefaultAzureCredential()

        self_harm_eval = SelfHarmEvaluator(azure_ai_project=azure_ai_project, credential=credential)
        self_harm_eval(
            query="What is the capital of France?",
            response="Paris",
        )
        # [END self_harm_evaluator]

        # [START sexual_evaluator]
        import os
        from azure.identity import DefaultAzureCredential
        from azure.ai.evaluation import SexualEvaluator

        azure_ai_project = {
            "subscription_id": os.environ.get("AZURE_SUBSCRIPTION_ID"),
            "resource_group_name": os.environ.get("AZURE_RESOURCE_GROUP_NAME"),
            "project_name": os.environ.get("AZURE_PROJECT_NAME"),
        }
        credential = DefaultAzureCredential()

        sexual_eval = SexualEvaluator(azure_ai_project=azure_ai_project, credential=credential)
        sexual_eval(
            query="What is the capital of France?",
            response="Paris",
        )
        # [END sexual_evaluator]

        # [START violence_evaluator]
        import os
        from azure.identity import DefaultAzureCredential
        from azure.ai.evaluation import ViolenceEvaluator

        azure_ai_project = {
            "subscription_id": os.environ.get("AZURE_SUBSCRIPTION_ID"),
            "resource_group_name": os.environ.get("AZURE_RESOURCE_GROUP_NAME"),
            "project_name": os.environ.get("AZURE_PROJECT_NAME"),
        }
        credential = DefaultAzureCredential()

        violence_eval = ViolenceEvaluator(azure_ai_project=azure_ai_project, credential=credential)
        violence_eval(
            query="What is the capital of France?",
            response="Paris",
        )
        # [END violence_evaluator]

        # [START f1_score_evaluator]
        from azure.ai.evaluation import F1ScoreEvaluator

        f1_evaluator = F1ScoreEvaluator()
        f1_evaluator(response="Lyon is the capital of France.", ground_truth="Paris is the capital of France.")
        # [END f1_score_evaluator]

        # [START fluency_evaluator]
        import os
        from azure.ai.evaluation import FluencyEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }

        fluency_evaluator = FluencyEvaluator(model_config=model_config)
        fluency_evaluator(response="Paris is the capital of France.")
        # [END fluency_evaluator]

        # [START gleu_score_evaluator]
        from azure.ai.evaluation import GleuScoreEvaluator

        gleu_evaluator = GleuScoreEvaluator()
        gleu_evaluator(response="Paris is the capital of France.", ground_truth="France's capital is Paris.")
        # [END gleu_score_evaluator]

        # [START groundedness_evaluator]
        import os
        from azure.ai.evaluation import GroundednessEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }

        groundedness_evaluator = GroundednessEvaluator(model_config=model_config)
        groundedness_evaluator(
            response="Paris is the capital of France.",
            context=(
                "France, a country in Western Europe, is known for its rich history and cultural heritage."
                "The city of Paris, located in the northern part of the country, serves as its capital."
                "Paris is renowned for its art, fashion, and landmarks such as the Eiffel Tower and the Louvre Museum."
            ),
        )
        # [END groundedness_evaluator]

        # [START meteor_score_evaluator]
        from azure.ai.evaluation import MeteorScoreEvaluator

        meteor_evaluator = MeteorScoreEvaluator(alpha=0.8)
        meteor_evaluator(response="Paris is the capital of France.", ground_truth="France's capital is Paris.")
        # [END meteor_score_evaluator]

        # [START protected_material_evaluator]
        import os
        from azure.identity import DefaultAzureCredential
        from azure.ai.evaluation import ProtectedMaterialEvaluator

        azure_ai_project = {
            "subscription_id": os.environ.get("AZURE_SUBSCRIPTION_ID"),
            "resource_group_name": os.environ.get("AZURE_RESOURCE_GROUP_NAME"),
            "project_name": os.environ.get("AZURE_PROJECT_NAME"),
        }
        credential = DefaultAzureCredential()

        protected_material_eval = ProtectedMaterialEvaluator(azure_ai_project=azure_ai_project, credential=credential)
        protected_material_eval(
            query="Write me a catchy song",
            response=(
                "You are the dancing queen, young and sweet, only seventeen."
                "Dancing queen, feel the beat from the tambourine, oh yeah."
            ),
        )
        # [END protected_material_evaluator]

        # [START qa_evaluator]
        import os
        from azure.ai.evaluation import QAEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }

        qa_eval = QAEvaluator(model_config=model_config)
        qa_eval(query="This's the color?", response="Black", ground_truth="gray", context="gray")
        # [END qa_evaluator]

        # [START relevance_evaluator]
        import os
        from azure.ai.evaluation import RelevanceEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }

        relevance_eval = RelevanceEvaluator(model_config=model_config)
        relevance_eval(
            query="What is the capital of Japan?",
            response="The capital of Japan is Tokyo.",
        )
        # [END relevance_evaluator]

        # [START retrieval_evaluator]
        import os
        from azure.ai.evaluation import RetrievalEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }

        retrieval_eval = RetrievalEvaluator(model_config=model_config)
        conversation = {
            "messages": [
                {
                    "content": "What is the capital of France?`''\"</>{}{{]",
                    "role": "user",
                    "context": "Customer wants to know the capital of France",
                },
                {"content": "Paris", "role": "assistant", "context": "Paris is the capital of France"},
                {
                    "content": "What is the capital of Hawaii?",
                    "role": "user",
                    "context": "Customer wants to know the capital of Hawaii",
                },
                {"content": "Honolulu", "role": "assistant", "context": "Honolulu is the capital of Hawaii"},
            ],
            "context": "Global context",
        }
        retrieval_eval(conversation=conversation)
        # [END retrieval_evaluator]

        # [START rouge_score_evaluator]
        from azure.ai.evaluation import RougeScoreEvaluator, RougeType

        rouge_evaluator = RougeScoreEvaluator(rouge_type=RougeType.ROUGE_4)
        rouge_evaluator(response="Paris is the capital of France.", ground_truth="France's capital is Paris.")
        # [END rouge_score_evaluator]

        # [START similarity_evaluator]
        import os
        from azure.ai.evaluation import SimilarityEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }

        similarity_eval = SimilarityEvaluator(model_config=model_config)
        similarity_eval(
            query="What is the capital of Japan?",
            response="The capital of Japan is Tokyo.",
            ground_truth="Tokyo is Japan's capital.",
        )
        # [END similarity_evaluator]

        # [START completeness_evaluator]
        import os
        from azure.ai.evaluation import CompletenessEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }

        completeness_eval = CompletenessEvaluator(model_config=model_config)
        completeness_eval(
            response="The capital of Japan is Tokyo.",
            ground_truth="Tokyo is Japan's capital.",
        )
        # [END completeness_evaluator]

        # [START task_adherence_evaluator]
        import os
        from azure.ai.evaluation import TaskAdherenceEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }

        task_adherence_evaluator = TaskAdherenceEvaluator(model_config=model_config)

        query = [{'role': 'system', 'content': 'You are a helpful customer service agent.'}, 
         {'role': 'user', 'content': [{'type': 'text', 'text': 'What is the status of my order #123?'}]}]

        response = [{'role': 'assistant', 'content': [{'type': 'tool_call', 'tool_call': {'id': 'tool_001', 'type': 'function', 'function': {'name': 'get_order', 'arguments': {'order_id': '123'}}}}]}, 
            {'role': 'tool', 'tool_call_id': 'tool_001', 'content': [{'type': 'tool_result', 'tool_result': '{ "order": { "id": "123", "status": "shipped" } }'}]}, 
            {'role': 'assistant', 'content': [{'type': 'text', 'text': 'Your order #123 has been shipped.'}]}]

        tool_definitions = [{'name': 'get_order', 'description': 'Get order details.', 'parameters': {'type': 'object', 'properties': {'order_id': {'type': 'string'}}}}]

        task_adherence_evaluator(
            query=query,
            response=response,
            tool_definitions=tool_definitions
        )
        # [END task_adherence_evaluator]

        # [START indirect_attack_evaluator]
        import os
        from azure.identity import DefaultAzureCredential
        from azure.ai.evaluation import IndirectAttackEvaluator

        azure_ai_project = {
            "subscription_id": os.environ.get("AZURE_SUBSCRIPTION_ID"),
            "resource_group_name": os.environ.get("AZURE_RESOURCE_GROUP_NAME"),
            "project_name": os.environ.get("AZURE_PROJECT_NAME"),
        }
        credential = DefaultAzureCredential()

        indirect_attack_eval = IndirectAttackEvaluator(azure_ai_project=azure_ai_project, credential=credential)
        indirect_attack_eval(
            query="What is the capital of France?",
            response="Paris",
        )
        # [END indirect_attack_evaluator]

        # [START groundedness_pro_evaluator]
        import os
        from azure.identity import DefaultAzureCredential
        from azure.ai.evaluation import GroundednessProEvaluator

        azure_ai_project = {
            "subscription_id": os.environ.get("AZURE_SUBSCRIPTION_ID"),
            "resource_group_name": os.environ.get("AZURE_RESOURCE_GROUP_NAME"),
            "project_name": os.environ.get("AZURE_PROJECT_NAME"),
        }
        credential = DefaultAzureCredential()

        groundedness_pro_eval = GroundednessProEvaluator(azure_ai_project=azure_ai_project, credential=credential)
        groundedness_pro_eval(
            query="What shape has 4 equilateral sides?",
            response="Rhombus",
            context="Rhombus is a shape with 4 equilateral sides.",
        )
        # [END groundedness_pro_evaluator]

        # [START tool_call_accuracy_evaluator]
        import os
        from azure.ai.evaluation import ToolCallAccuracyEvaluator

        model_config = {
            "azure_endpoint": os.environ.get("AZURE_OPENAI_ENDPOINT"),
            "api_key": os.environ.get("AZURE_OPENAI_KEY"),
            "azure_deployment": os.environ.get("AZURE_OPENAI_DEPLOYMENT"),
        }

        tool_call_accuracy_evaluator = ToolCallAccuracyEvaluator(model_config=model_config)
        tool_call_accuracy_evaluator(
            query="How is the weather in New York?",
            response="The weather in New York is sunny.",
            tool_calls={
                "type": "tool_call",
                "tool_call": {
                    "id": "call_eYtq7fMyHxDWIgeG2s26h0lJ",
                    "type": "function",
                    "function": {
                        "name": "fetch_weather",
                        "arguments": {
                            "location": "New York"
                        }
                    }
                }
            },
            tool_definitions={
                "id": "fetch_weather",
                "name": "fetch_weather",
                "description": "Fetches the weather information for the specified location.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The location to fetch weather for."
                        }
                    }
                }
            }
        )
        # [END tool_call_accuracy_evaluator]

        # [START document_retrieval_evaluator]
        from azure.ai.evaluation import DocumentRetrievalEvaluator

        retrieval_ground_truth = [
            {
                "document_id": "1",
                "query_relevance_judgement": 4
            },
            {
                "document_id": "2",
                "query_relevance_judgement": 2
            },
            {
                "document_id": "3",
                "query_relevance_judgement": 3
            },
            {
                "document_id": "4",
                "query_relevance_judgement": 1
            },
            {
                "document_id": "5",
                "query_relevance_judgement": 0
            },
        ]

        retrieved_documents = [
            {
                "document_id": "2",
                "query_relevance_judgement": 45.1
            },
            {
                "document_id": "6",
                "query_relevance_judgement": 35.8
            },
            {
                "document_id": "3",
                "query_relevance_judgement": 29.2
            },
            {
                "document_id": "5",
                "query_relevance_judgement": 25.4
            },
            {
                "document_id": "7",
                "query_relevance_judgement": 18.8
            },
        ]

        document_retrieval_evaluator = DocumentRetrievalEvaluator()
        document_retrieval_evaluator(retrieval_ground_truth=retrieval_ground_truth, retrieved_documents=retrieved_documents)        
        # [END document_retrieval_evaluator]


if __name__ == "__main__":
    from dotenv import load_dotenv
    load_dotenv()

    print("Loading samples in evaluation_samples_evaluate.py")
    sample = EvaluationEvaluateSamples()
    print("Samples loaded successfully!")
    print("Running samples in evaluation_samples_evaluate.py")
    sample.evaluation_evaluate_classes_methods()
    print("Samples ran successfully!")



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Callable, Final


try:
    from promptflow._cli._pf._service import stop_service as _stop_service
except ImportError:

    def _stop_service() -> None:
        pass


stop_service: Final[Callable[[], None]] = _stop_service



# Serialization and deserialization helper functions to make test-writing easier

import json
from datetime import datetime

from azure.ai.evaluation._converters._models import ToolCall


# Breaking changes introduced in newer version of the agents SDK
# Models have been moved, so try a few different locations
try:
    from azure.ai.projects.models import (
        RunStepCodeInterpreterToolCall,
        RunStepCodeInterpreterToolCallDetails,
        RunStepFileSearchToolCall,
        RunStepFileSearchToolCallResults,
        RunStepFileSearchToolCallResult,
        FileSearchRankingOptions,
        RunStepBingGroundingToolCall,
        ThreadRun,
    )
except ImportError:
    pass
try:
    from azure.ai.agents.models import (
        RunStepCodeInterpreterToolCall,
        RunStepCodeInterpreterToolCallDetails,
        RunStepFileSearchToolCall,
        RunStepFileSearchToolCallResults,
        RunStepFileSearchToolCallResult,
        FileSearchRankingOptions,
        RunStepBingGroundingToolCall,
        ThreadRun,
    )
except ImportError:
    pass



class ToolDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if 'completed' in obj and 'created' in obj and 'details' in obj:
            return ToolCall(
                created=datetime.fromisoformat(obj['created']),
                completed=datetime.fromisoformat(obj['completed']),
                details=self.decode_details(obj['details'])
            )
        return obj

    def decode_details(self, details):
        if 'id' in details and 'type' in details:
            if details['type'] == 'code_interpreter':
                return RunStepCodeInterpreterToolCall(
                    id=details['id'],
                    code_interpreter=RunStepCodeInterpreterToolCallDetails(
                        input=details['code_interpreter']['input'],
                        outputs=details['code_interpreter']['outputs']
                    )
                )
            elif details['type'] == 'file_search':
                return RunStepFileSearchToolCall(
                    id=details['id'],
                    file_search=RunStepFileSearchToolCallResults(
                        results=[
                            RunStepFileSearchToolCallResult(
                                file_name=result['file_name'],
                                file_id=result['file_id'],
                                score=result['score'],
                                content=result['content']
                            ) for result in details['file_search']['results']
                        ],
                        ranking_options=FileSearchRankingOptions(
                            ranker=details['file_search']['ranking_options']['ranker'],
                            score_threshold=details['file_search']['ranking_options']['score_threshold']
                        )
                    )
                )
            elif details['type'] == 'bing_grounding':
                return RunStepBingGroundingToolCall(
                    id=details['id'],
                    bing_grounding=details['bing_grounding']
                )
        return details

class ToolEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, ToolCall):
            return {
                "completed": obj.completed,
                "created": obj.created,
                "details": obj.details
            }
        if isinstance(obj, RunStepCodeInterpreterToolCall):
            return {
                "id": obj.id,
                # "type": obj.type,
                "code_interpreter": obj.code_interpreter
            }
        if isinstance(obj, RunStepCodeInterpreterToolCallDetails):
            return {
                "input": obj.input,
                "outputs": obj.outputs
            }
        if isinstance(obj, RunStepFileSearchToolCall):
            return {
                "id": obj.id,
                # "type": obj.type,
                "file_search": obj.file_search
            }
        if isinstance(obj, RunStepFileSearchToolCallResults):
            return {
                "results": obj.results,
                "ranking_options": obj.ranking_options
            }
        if isinstance(obj, RunStepFileSearchToolCallResult):
            return {
                "file_name": obj.file_name,
                "file_id": obj.file_id,
                "score": obj.score,
                "content": obj.content
            }
        if isinstance(obj, FileSearchRankingOptions):
            return {
                "ranker": obj.ranker,
                "score_threshold": obj.score_threshold
            }
        if isinstance(obj, RunStepBingGroundingToolCall):
            return {
                "id": obj.id,
                # "type": obj.type,
                "bing_grounding": obj.bing_grounding
            }
        return super().default(obj)

class ThreadRunDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if "id" in obj and "thread_id" in obj:
            return ThreadRun(
                id=obj["id"],
                thread_id=obj["thread_id"],
                agent_id=obj["assistant_id"],
                status=obj["status"],
                required_action=obj.get("required_action"),
                last_error=obj.get("last_error"),
                model=obj["model"],
                instructions=obj["instructions"],
                tools=obj["tools"],
                created_at=datetime.fromtimestamp(obj["created_at"]),
                expires_at=datetime.fromtimestamp(obj["expires_at"]) if obj.get("expires_at") else None,
                started_at=datetime.fromtimestamp(obj["started_at"]) if obj.get("started_at") else None,
                completed_at=datetime.fromtimestamp(obj["completed_at"]) if obj.get("completed_at") else None,
                cancelled_at=datetime.fromtimestamp(obj["cancelled_at"]) if obj.get("cancelled_at") else None,
                failed_at=datetime.fromtimestamp(obj["failed_at"]) if obj.get("failed_at") else None,
                incomplete_details=obj.get("incomplete_details"),
                usage=obj.get("usage"),
                temperature=obj.get("temperature"),
                top_p=obj.get("top_p"),
                max_prompt_tokens=obj["max_prompt_tokens"],
                max_completion_tokens=obj["max_completion_tokens"],
                truncation_strategy=obj["truncation_strategy"],
                tool_choice=obj["tool_choice"],
                response_format=obj["response_format"],
                metadata=obj["metadata"],
                tool_resources=obj.get("tool_resources"),
                parallel_tool_calls=obj["parallel_tool_calls"],
            )
        return obj


class ThreadRunEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, ThreadRun):
            return {
                "id": obj.id,
                "object": obj.object,
                "thread_id": obj.thread_id,
                "assistant_id": obj.assistant_id,
                "status": obj.status,
                "required_action": obj.required_action,
                "last_error": obj.last_error,
                "model": obj.model,
                "instructions": obj.instructions,
                "tools": obj.tools,
                "created_at": int(obj.created_at.timestamp()),
                "expires_at": int(obj.expires_at.timestamp()) if obj.expires_at else None,
                "started_at": int(obj.started_at.timestamp()) if obj.started_at else None,
                "completed_at": int(obj.completed_at.timestamp()) if obj.completed_at else None,
                "cancelled_at": int(obj.cancelled_at.timestamp()) if obj.cancelled_at else None,
                "failed_at": int(obj.failed_at.timestamp()) if obj.failed_at else None,
                "incomplete_details": obj.incomplete_details,
                "usage": obj.usage,
                "temperature": obj.temperature,
                "top_p": obj.top_p,
                "max_prompt_tokens": obj.max_prompt_tokens,
                "max_completion_tokens": obj.max_completion_tokens,
                "truncation_strategy": obj.truncation_strategy,
                "tool_choice": obj.tool_choice,
                "response_format": obj.response_format,
                "metadata": obj.metadata,
                "tool_resources": obj.tool_resources,
                "parallel_tool_calls": obj.parallel_tool_calls,
            }
        return super().default(obj)


import json
import logging
import os
import pathlib
from unittest.mock import MagicMock, patch

import pytest
from devtools_testutils import is_live
from ci_tools.variables import in_ci

from azure.ai.evaluation import F1ScoreEvaluator
from azure.ai.evaluation._evaluate import _utils as ev_utils
from azure.ai.evaluation._evaluate._eval_run import EvalRun
from azure.ai.evaluation._evaluate._evaluate import evaluate
from azure.ai.evaluation._azure._clients import LiteMLClient


@pytest.fixture
def data_file():
    data_path = os.path.join(pathlib.Path(__file__).parent.resolve(), "data")
    return os.path.join(data_path, "evaluate_test_data.jsonl")


@pytest.fixture
def questions_answers_file():
    data_path = os.path.join(pathlib.Path(__file__).parent.resolve(), "data")
    return os.path.join(data_path, "questions_answers.jsonl")


@pytest.fixture
def questions_file():
    data_path = os.path.join(pathlib.Path(__file__).parent.resolve(), "data")
    return os.path.join(data_path, "questions.jsonl")


def _get_tracking_uri(azure_ml_client: LiteMLClient, project_scope: dict) -> str:
    return azure_ml_client.workspace_get_info(project_scope["project_name"]).ml_flow_tracking_uri or ""


@pytest.mark.usefixtures("model_config", "recording_injection", "project_scope", "recorded_test")
class TestMetricsUpload(object):
    """End to end tests to check how the metrics were uploaded to cloud."""

    # NOTE:
    # If you are re-recording the tests, remember to disable Promptflow telemetry from the command line using:
    # pf config set telemetry.enabled=false
    # Otherwise you will capture telemetry requests in the recording which will cause test playback failures.

    def _assert_no_errors_for_module(self, records, module_names):
        """Check there are no errors in the log."""
        error_messages = []
        if records:
            error_messages = [
                lg_rec.message
                for lg_rec in records
                if lg_rec.levelno == logging.WARNING and (lg_rec.name in module_names)
            ]
            assert not error_messages, "\n".join(error_messages)

    @pytest.mark.azuretest
    def test_writing_to_run_history(self, caplog, project_scope, azure_ml_client: LiteMLClient):
        """Test logging data to RunHistory service."""
        logger = logging.getLogger(EvalRun.__module__)
        # All loggers, having promptflow. prefix will have "promptflow" logger
        # as a parent. This logger does not propagate the logs and cannot be
        # captured by caplog. Here we will skip this logger to capture logs.
        logger.parent = logging.root
        # Just for sanity check let us make sure that the logging actually works
        mock_response = MagicMock()
        mock_response.status_code = 418
        with EvalRun(
            run_name="test",
            tracking_uri=_get_tracking_uri(azure_ml_client, project_scope),
            subscription_id=project_scope["subscription_id"],
            group_name=project_scope["resource_group_name"],
            workspace_name=project_scope["project_name"],
            management_client=azure_ml_client,
        ) as ev_run:
            with patch(
                "azure.ai.evaluation._evaluate._eval_run.EvalRun.request_with_retry", return_value=mock_response
            ):
                ev_run.write_properties_to_run_history({"test": 42})
                assert any(
                    lg_rec.levelno == logging.ERROR for lg_rec in caplog.records
                ), "The error log was not captured!"
            caplog.clear()
            ev_run.write_properties_to_run_history({"test": 42})
        self._assert_no_errors_for_module(caplog.records, [EvalRun.__module__])

    @pytest.mark.azuretest
    def test_logging_metrics(self, caplog, project_scope, azure_ml_client):
        """Test logging metrics."""
        logger = logging.getLogger(EvalRun.__module__)
        # All loggers, having promptflow. prefix will have "promptflow" logger
        # as a parent. This logger does not propagate the logs and cannot be
        # captured by caplog. Here we will skip this logger to capture logs.
        logger.parent = logging.root
        with EvalRun(
            run_name="test",
            tracking_uri=_get_tracking_uri(azure_ml_client, project_scope),
            subscription_id=project_scope["subscription_id"],
            group_name=project_scope["resource_group_name"],
            workspace_name=project_scope["project_name"],
            management_client=azure_ml_client,
        ) as ev_run:
            mock_response = MagicMock()
            mock_response.status_code = 418
            with patch(
                "azure.ai.evaluation._evaluate._eval_run.EvalRun.request_with_retry", return_value=mock_response
            ):
                ev_run.log_metric("f1", 0.54)
                assert any(
                    lg_rec.levelno == logging.WARNING for lg_rec in caplog.records
                ), "The error log was not captured!"
            caplog.clear()
            ev_run.log_metric("f1", 0.54)
        self._assert_no_errors_for_module(caplog.records, EvalRun.__module__)

    @pytest.mark.azuretest
    @pytest.mark.parametrize("config_name", ["sas", "none"])
    def test_log_artifact(self, project_scope, azure_cred, datastore_project_scopes, caplog, tmp_path, config_name):
        """Test uploading artifact to the service."""
        logger = logging.getLogger(EvalRun.__module__)
        # All loggers, having promptflow. prefix will have "promptflow" logger
        # as a parent. This logger does not propagate the logs and cannot be
        # captured by caplog. Here we will skip this logger to capture logs.
        logger.parent = logging.root

        project_scope = datastore_project_scopes[config_name]
        azure_ml_client = LiteMLClient(
            subscription_id=project_scope["subscription_id"],
            resource_group=project_scope["resource_group_name"],
            logger=logger,
            credential=azure_cred,
        )

        with EvalRun(
            run_name="test",
            tracking_uri=_get_tracking_uri(azure_ml_client, project_scope),
            subscription_id=project_scope["subscription_id"],
            group_name=project_scope["resource_group_name"],
            workspace_name=project_scope["project_name"],
            management_client=azure_ml_client,
        ) as ev_run:
            mock_response = MagicMock()
            mock_response.status_code = 418
            with open(os.path.join(tmp_path, EvalRun.EVALUATION_ARTIFACT), "w") as fp:
                json.dump({"f1": 0.5}, fp)
            os.makedirs(os.path.join(tmp_path, "internal_dir"), exist_ok=True)
            with open(os.path.join(tmp_path, "internal_dir", "test.json"), "w") as fp:
                json.dump({"internal_f1": 0.6}, fp)
            with patch(
                "azure.ai.evaluation._evaluate._eval_run.EvalRun.request_with_retry", return_value=mock_response
            ):
                ev_run.log_artifact(tmp_path)
                assert any(
                    lg_rec.levelno == logging.WARNING for lg_rec in caplog.records
                ), "The error log was not captured!"
            caplog.clear()
            ev_run.log_artifact(tmp_path)
        self._assert_no_errors_for_module(caplog.records, EvalRun.__module__)

    @pytest.mark.performance_test
    @pytest.mark.skipif(
        in_ci(),
        reason="There is some weird JSON serialiazation issue that only appears in CI where a \n becomes a \r\n",
    )
    def test_e2e_run_target_fn(self, caplog, project_scope, questions_answers_file, monkeypatch, azure_cred):
        """Test evaluation run logging."""
        # Afer re-recording this test, please make sure, that the cassette contains the POST
        # request ending by 00000/rundata and it has status 200.
        # Also make sure that the cosmos request ending by workspaces/00000/TraceSessions
        # and log metric call anding on /mlflow/runs/log-metric are also present.
        # pytest-cov generates coverage files, which are being uploaded. When recording tests,
        # make sure to enable coverage, check that .coverage.sanitized-suffix is present
        # in the cassette.

        # We cannot define target in this file as pytest will load
        # all modules in test folder and target_fn will be imported from the first
        # module named test_evaluate and it will be a different module in unit test
        # folder. By keeping function in separate file we guarantee, it will be loaded
        # from there.
        logger = logging.getLogger(EvalRun.__module__)
        # Switch off tracing as it is running in the second thread, wile
        # thread pool executor is not compatible with VCR.py.
        if not is_live():
            try:
                from promptflow.tracing import _start_trace
                monkeypatch.setattr(_start_trace, "_is_devkit_installed", lambda: False)
            except ImportError:
                pass
        # All loggers, having promptflow. prefix will have "promptflow" logger
        # as a parent. This logger does not propagate the logs and cannot be
        # captured by caplog. Here we will skip this logger to capture logs.
        logger.parent = logging.root
        from .target_fn import target_fn

        f1_score_eval = F1ScoreEvaluator()
        evaluate(
            data=questions_answers_file,
            target=target_fn,
            evaluators={"f1": f1_score_eval},
            azure_ai_project=project_scope,
            credential=azure_cred,
        )
        self._assert_no_errors_for_module(caplog.records, (ev_utils.__name__, EvalRun.__module__))

    @pytest.mark.performance_test
    @pytest.mark.skipif(
        in_ci(),
        reason="There is some weird JSON serialiazation issue that only appears in CI where a \n becomes a \r\n",
    )
    def test_e2e_run(self, caplog, project_scope, questions_answers_file, monkeypatch, azure_cred):
        """Test evaluation run logging."""
        # Afer re-recording this test, please make sure, that the cassette contains the POST
        # request ending by /BulkRuns/create.
        # Also make sure that the cosmos request ending by workspaces/00000/TraceSessions
        # is also present.
        # pytest-cov generates coverage files, which are being uploaded. When recording tests,
        # make sure to enable coverage, check that .coverage.sanitized-suffix is present
        # in the cassette.
        logger = logging.getLogger(EvalRun.__module__)
        # All loggers, having promptflow. prefix will have "promptflow" logger
        # as a parent. This logger does not propagate the logs and cannot be
        # captured by caplog. Here we will skip this logger to capture logs.
        logger.parent = logging.root
        # Switch off tracing as it is running in the second thread, wile
        # thread pool executor is not compatible with VCR.py.
        if not is_live():
            try:
                from promptflow.tracing import _start_trace
                monkeypatch.setattr(_start_trace, "_is_devkit_installed", lambda: False)
            except ImportError:
                pass
        f1_score_eval = F1ScoreEvaluator()
        evaluate(
            data=questions_answers_file,
            evaluators={"f1": f1_score_eval},
            azure_ai_project=project_scope,
            credential=azure_cred,
        )
        self._assert_no_errors_for_module(caplog.records, (ev_utils.__name__, EvalRun.__module__))



import os
from unittest.mock import MagicMock

import pytest
from azure.ai.evaluation._legacy._adapters.client import PFClient
from azure.ai.evaluation._legacy._adapters._check import MISSING_LEGACY_SDK

from azure.ai.evaluation._constants import PF_BATCH_TIMEOUT_SEC, PF_BATCH_TIMEOUT_SEC_DEFAULT
from azure.ai.evaluation._evaluate._batch_run import CodeClient, EvalRunContext, ProxyClient
from azure.ai.evaluation._user_agent import USER_AGENT


@pytest.fixture
def code_client_mock():
    return MagicMock(spec=CodeClient)


@pytest.fixture
def pf_client_mock():
    return MagicMock(spec=PFClient)


@pytest.mark.unittest
class TestEvalRunContext:
    @pytest.mark.skipif(MISSING_LEGACY_SDK, reason="This test has a promptflow dependency")
    def test_with_codeclient(self, mocker, code_client_mock):
        mock_append_user_agent = mocker.patch(
            "promptflow._utils.user_agent_utils.ClientUserAgentUtil.append_user_agent"
        )
        mock_inject_openai_api = mocker.patch("promptflow.tracing._integrations._openai_injector.inject_openai_api")
        mock_recover_openai_api = mocker.patch("promptflow.tracing._integrations._openai_injector.recover_openai_api")

        with EvalRunContext(code_client_mock):
            # TODO: Failed to mock inject_openai_api and recover_openai_api for some reason.
            # Need to investigate further.
            # mock_inject_openai_api.assert_called_once()
            # mock_recover_openai_api.assert_called_once()
            print(f"mock_inject_openai_api.call_count: {mock_inject_openai_api.call_count}")
            print(f"mock_recover_openai_api.call_count: {mock_recover_openai_api.call_count}")
            pass

            mock_append_user_agent.assert_called_once_with(USER_AGENT)

    @pytest.mark.skipif(MISSING_LEGACY_SDK, reason="This test has a promptflow dependency")
    def test_with_pfclient(self, mocker, pf_client_mock):
        mock_append_user_agent = mocker.patch(
            "promptflow._utils.user_agent_utils.ClientUserAgentUtil.append_user_agent"
        )
        mock_inject_openai_api = mocker.patch("promptflow.tracing._integrations._openai_injector.inject_openai_api")
        mock_recover_openai_api = mocker.patch("promptflow.tracing._integrations._openai_injector.recover_openai_api")

        with EvalRunContext(pf_client_mock):
            mock_append_user_agent.assert_not_called()
            mock_inject_openai_api.assert_not_called()
            pass

        mock_recover_openai_api.assert_not_called()

    def test_batch_timeout_default(self):
        before_timeout = os.environ.get(PF_BATCH_TIMEOUT_SEC)
        assert before_timeout is None

        with EvalRunContext(ProxyClient()):
            during_timeout = int(os.environ.get(PF_BATCH_TIMEOUT_SEC))
            assert during_timeout == PF_BATCH_TIMEOUT_SEC_DEFAULT

        # Default timeout should be reset after exiting EvalRunContext
        after_timeout = os.environ.get(PF_BATCH_TIMEOUT_SEC)
        assert after_timeout is None

    def test_batch_timeout_custom(self):
        custom_timeout = 1000
        os.environ[PF_BATCH_TIMEOUT_SEC] = str(custom_timeout)

        with EvalRunContext(ProxyClient()):
            during_timeout = int(os.environ.get(PF_BATCH_TIMEOUT_SEC))
            assert during_timeout == custom_timeout

        # Custom timeouts should not be reset after exiting EvalRunContext
        after_timeout = int(os.environ.get(PF_BATCH_TIMEOUT_SEC))
        assert after_timeout == custom_timeout



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import logging
from typing import Union, Any, Dict
from azure.core.credentials import AzureKeyCredential, TokenCredential
from azure.ai.evaluation._common.onedp import AIProjectClient as RestEvaluationServiceClient
from azure.ai.evaluation._common.onedp.models import (PendingUploadRequest, PendingUploadType, EvaluationResult,
                                                      ResultType, AssetCredentialRequest, EvaluationUpload, InputDataset, RedTeamUpload)
from azure.storage.blob import ContainerClient
from .utils import upload

LOGGER = logging.getLogger(__name__)

class EvaluationServiceOneDPClient:

    def __init__(self, endpoint: str, credential: Union[AzureKeyCredential, "TokenCredential"], **kwargs: Any) -> None:
        self.rest_client = RestEvaluationServiceClient(
            endpoint=endpoint,
            credential=credential,
            **kwargs,
        )

    def create_evaluation_result(
            self, *, name: str, path: str, version=1, metrics: Dict[str, int]=None, result_type: ResultType=ResultType.EVALUATION, **kwargs) -> EvaluationResult:
        """Create and upload evaluation results to Azure evaluation service.

        This method uploads evaluation results from a local path to Azure Blob Storage
        and registers them with the evaluation service. The process involves:
        1. Starting a pending upload with the evaluation service
        2. Getting a SAS token for the blob container
        3. Uploading the local evaluation results to the blob container
        4. Creating a version record for the evaluation results

        :param name: The name to identify the evaluation results
        :type name: str
        :param path: The local path to the evaluation results file or directory
        :type path: str
        :param version: The version number for the evaluation results, defaults to 1
        :type version: int, optional
        :param metrics: Metrics to be added to evaluation result
        :type metrics: Dict[str, int], optional
        :param result_type: Evaluation Result Type to create
        :type result_type: ResultType, optional
        :param kwargs: Additional keyword arguments to pass to the underlying API calls
        :return: The response from creating the evaluation result version
        :rtype: EvaluationResult
        :raises: Various exceptions from the underlying API calls or upload process
        """

        LOGGER.debug(f"Creating evaluation result for {name} with version {version} type {result_type} from path {path}")
        start_pending_upload_response = self.rest_client.evaluation_results.start_pending_upload(
            name=name,
            version=version,
            body=PendingUploadRequest(pending_upload_type=PendingUploadType.TEMPORARY_BLOB_REFERENCE),
            **kwargs
        )

        LOGGER.debug(f"Uploading {path} to {start_pending_upload_response.blob_reference_for_consumption.blob_uri}")
        with ContainerClient.from_container_url(
            start_pending_upload_response.blob_reference_for_consumption.credential.sas_uri) as container_client:
            upload(path=path, container_client=container_client, logger=LOGGER)

        LOGGER.debug(f"Creating evaluation result version for {name} with version {version}")
        create_version_response = self.rest_client.evaluation_results.create_or_update_version(
            body=EvaluationResult(
                blob_uri=start_pending_upload_response.blob_reference_for_consumption.blob_uri,
                result_type=result_type,
                name=name,
                version=version,
                metrics=metrics,
            ),
            name=name,
            version=version,
            **kwargs
        )

        return create_version_response

    def start_evaluation_run(self, *, evaluation: EvaluationUpload, **kwargs) -> EvaluationUpload:
        """Start a new evaluation run in the Azure evaluation service.

        This method creates a new evaluation run with the provided configuration details.

        :param evaluation: The evaluation configuration to upload
        :type evaluation: EvaluationUpload
        :param kwargs: Additional keyword arguments to pass to the underlying API calls
        :return: The created evaluation run object
        :rtype: EvaluationUpload
        :raises: Various exceptions from the underlying API calls
        """
        upload_run_response = self.rest_client.evaluations.upload_run(
            evaluation=evaluation,
            **kwargs
        )

        return upload_run_response

    def update_evaluation_run(self, *, name: str, evaluation: EvaluationUpload, **kwargs) -> EvaluationUpload:
        """Update an existing evaluation run in the Azure evaluation service.

        This method updates an evaluation run with new information such as status changes,
        result references, or other metadata.

        :param name: The identifier of the evaluation run to update
        :type name: str
        :param evaluation: The updated evaluation configuration
        :type evaluation: EvaluationUpload
        :param kwargs: Additional keyword arguments to pass to the underlying API calls
        :return: The updated evaluation run object
        :rtype: EvaluationUpload
        :raises: Various exceptions from the underlying API calls
        """
        update_run_response = self.rest_client.evaluations.upload_update_run(
            name=name,
            evaluation=evaluation,
            **kwargs
        )

        return update_run_response

    def start_red_team_run(self, *, red_team: RedTeamUpload, **kwargs):
        """Start a new red team run in the Azure evaluation service.

        This method creates a new red team run with the provided configuration details.

        :param red_team: The red team configuration to upload
        :type red_team: ~azure.ai.evaluation._common.onedp.models.RedTeamUpload
        :param kwargs: Additional keyword arguments to pass to the underlying API calls
        :return: The created red team run object
        :rtype: ~azure.ai.evaluation._common.onedp.models.RedTeamUpload
        :raises: Various exceptions from the underlying API calls
        """
        upload_run_response = self.rest_client.red_teams.upload_run(
            redteam=red_team,
            **kwargs
        )

        return upload_run_response

    def update_red_team_run(self, *, name: str, red_team: RedTeamUpload, **kwargs):
        """Update an existing red team run in the Azure evaluation service.

        This method updates a red team run with new information such as status changes,
        result references, or other metadata.

        :param name: The identifier of the red team run to update
        :type name: str
        :param red_team: The updated red team configuration
        :type red_team: ~azure.ai.evaluation._common.onedp.models.RedTeamUpload
        :param kwargs: Additional keyword arguments to pass to the underlying API calls
        :return: The updated red team run object
        :rtype: ~azure.ai.evaluation._common.onedp.models.RedTeamUpload
        :raises: Various exceptions from the underlying API calls
        """
        update_run_response = self.rest_client.red_teams.upload_update_run(
            name=name,
            redteam=red_team,
            **kwargs
        )

        return update_run_response


# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import math
import os
import logging
import re
from typing import Dict, List, Union, TypeVar, cast
from typing_extensions import overload, override
from azure.ai.evaluation._evaluators._common import PromptyEvaluatorBase
from azure.ai.evaluation._common.utils import remove_optional_singletons, parse_quality_evaluator_reason_score
from azure.ai.evaluation._exceptions import ErrorBlame, ErrorCategory, ErrorTarget, EvaluationException
from azure.ai.evaluation._common.constants import PROMPT_BASED_REASON_EVALUATORS
from azure.ai.evaluation._common._experimental import experimental

logger = logging.getLogger(__name__)

T_EvalValue = TypeVar("T_EvalValue")

@experimental
class ToolCallAccuracyEvaluator(PromptyEvaluatorBase[Union[str, float]]):
    """The Tool Call Accuracy evaluator assesses how accurately an AI uses tools by examining:
        - Relevance to the conversation
        - Parameter correctness according to tool definitions
        - Parameter value extraction from the conversation

    The evaluator uses a binary scoring system (0 or 1):
        - Score 0: The tool call is irrelevant or contains information not in the conversation/definition
        - Score 1: The tool call is relevant with properly extracted parameters from the conversation

    This evaluation focuses on measuring whether tool calls meaningfully contribute to addressing
    user needs while properly following tool definitions and using information present in the
    conversation history.

    :param model_config: Configuration for the Azure OpenAI model.
    :type model_config: Union[~azure.ai.evaluation.AzureOpenAIModelConfiguration,
        ~azure.ai.evaluation.OpenAIModelConfiguration]

    .. admonition:: Example:

        .. literalinclude:: ../samples/evaluation_samples_evaluate.py
            :start-after: [START tool_call_accuracy_evaluator]
            :end-before: [END tool_call_accuracy_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call a ToolCallAccuracyEvaluator.

    .. admonition:: Example using Azure AI Project URL:
        
        .. literalinclude:: ../samples/evaluation_samples_evaluate_fdp.py
            :start-after: [START tool_call_accuracy_evaluator]
            :end-before: [END tool_call_accuracy_evaluator]
            :language: python
            :dedent: 8
            :caption: Initialize and call ToolCallAccuracyEvaluator using Azure AI Project URL in the following format 
                https://{resource_name}.services.ai.azure.com/api/projects/{project_name}

    .. note::

        To align with our support of a diverse set of models, an output key without the `gpt_` prefix has been added.
        To maintain backwards compatibility, the old key with the `gpt_` prefix is still be present in the output;
        however, it is recommended to use the new key moving forward as the old key will be deprecated in the future.
    """

    _PROMPTY_FILE = "tool_call_accuracy.prompty"
    _RESULT_KEY = "tool_call_accurate"
    _AGGREGATE_RESULT_KEY = "tool_call_accuracy"

    _MAX_TOOL_CALL_ACCURACY_SCORE = 1.0
    _MIN_TOOL_CALL_ACCURACY_SCORE = 0.0
    _DEFAULT_TOOL_CALL_ACCURACY_SCORE = 0.8

    id = "id"
    """Evaluator identifier, experimental and to be used only with evaluation in cloud."""

    @override
    def __init__(self, model_config, *,
                 threshold=_DEFAULT_TOOL_CALL_ACCURACY_SCORE,
                 **kwargs):
        current_dir = os.path.dirname(__file__)
        prompty_path = os.path.join(current_dir, self._PROMPTY_FILE)
        self.threshold = threshold
        super().__init__(model_config=model_config, prompty_file=prompty_path,
                         result_key=self._RESULT_KEY,
                         **kwargs)

    @overload
    def __call__(
        self,
        *,
        query: Union[str, List[dict]],
        tool_definitions: Union[dict, List[dict]],
        tool_calls: Union[dict, List[dict]]  = None,
        response: Union[str, List[dict]] = None
    ) -> Dict[str, Union[str, float]]:
        """
        Evaluate tool call accuracy. Accepts a query, tool definitions, and tool calls for evaluation.

        :keyword query: Query or Chat history up to the message that has the tool call being evaluated.
        :paramtype query: Union[str, List[dict]]
        :keyword tool_definitions: List of tool definitions whose calls are being evaluated.
        :paramtype tool_definitions: Union[dict, List[dict]]
        :keyword tool_calls: Optional List of tool calls to evaluate. If not provided response should be provided and should have
            tool call(s) in it.
        :paramtype tool_calls: Union[dict, List[dict]]
        :keyword response: Optional response to be evaluated alongside the tool calls.
            If provided all tool calls in response will be evaluated when tool_calls parameter is not provided.
            If provided and tool_calls parameter is provided, only the tool calls in tool_calls parameter will be evaluated.
                If response has extra tool calls they will not be evaluated, response will be used to extract any tool calls that are needed for evaluating a certain tool call.
            Recommended to provide it when there are tool calls that depend on output of a previous tool call.
        :paramtype response: Union[str, List[dict]]
        :return: The tool selection evaluation results.
        :rtype: Dict[str, Union[str, float]]
        """

    def _convert_kwargs_to_eval_input(self, **kwargs):
        """Convert an arbitrary input into a list of inputs for evaluators.
        It is assumed that evaluators generally make use of their inputs in one of two ways.
        Either they receive a collection of keyname inputs that are all single values
        (like a query and response), or they receive conversation that iss a list of dictionary
        values.

        The self._singleton_inputs list assigned during initialization is used to find and extract
        singleton keywords, and self._allow_conversation_input is used to determine if a conversation
        is a valid input.

        If both conversations and singletons are allowed, the function will raise an exception if both
        are inputted.

        This function must be overridden by child classes IF they need to both a conversation and
        other inputs to be passed in.

        :keyword kwargs: The inputs to convert.
        :type kwargs: Dict
        :return: A list of arbitrary values that are valid inputs for this evaluator's do_eval function.
        :rtype: List
        """
        # TODO add warning that only tool calls of type function are supported
        # Collect inputs
        tool_calls = kwargs.get("tool_calls", None)
        tool_definitions = kwargs.get("tool_definitions")
        query = kwargs.get("query", None)
        response = kwargs.get("response", None)

        if response is None and tool_calls is None:
            raise EvaluationException(
                message="Either response or tool_calls must be provided.",
                blame=ErrorBlame.USER_ERROR,
                category=ErrorCategory.MISSING_FIELD,
                target=ErrorTarget.TOOL_CALL_ACCURACY_EVALUATOR,
            )

        if tool_definitions is None:
            raise EvaluationException(
                message="Tool definitions must be provided.",
                blame=ErrorBlame.USER_ERROR,
                category=ErrorCategory.MISSING_FIELD,
                target=ErrorTarget.TOOL_CALL_ACCURACY_EVALUATOR,
            )

        # TODO : Support classes that represents tool calls, messages etc once client side definitions are available
        if tool_calls is None:
            # Extract tool calls from response if not provided
            tool_calls = []
            if isinstance(response, list):
                for message in response:
                    if message.get("role") == "assistant":
                        tool_calls.extend([content for content in message.get("content")
                                        if content.get("type") == "tool_call"])
            if len(tool_calls) == 0:
                raise EvaluationException(
                    message="response does not have tool calls. Either provide tool_calls or response with tool calls.",
                    blame=ErrorBlame.USER_ERROR,
                    category=ErrorCategory.MISSING_FIELD,
                    target=ErrorTarget.TOOL_CALL_ACCURACY_EVALUATOR,
                )

        if not isinstance(tool_calls, list):
            tool_calls = [tool_calls]

        if not isinstance(tool_definitions, list):
            tool_definitions = [tool_definitions]

        eval_inputs = []
        # TODO : When evaluating an agent tool that depends on the output of a previous tool call,
        # we need to provide the output of the previous tool call as part of messages.
        for tool_call in tool_calls:
            if isinstance(tool_call, dict) and tool_call.get("type") == "tool_call":  # TODO assuming dict here but it can be a class
                function_name = tool_call.get("name")
                tool_definition = [tool for tool in tool_definitions if tool.get("name") == function_name]
                if len(tool_definition) > 0:
                    tool_definition = tool_definition
                else:
                    raise EvaluationException(
                        message="Tool definition not found",
                        blame=ErrorBlame.USER_ERROR,
                        category=ErrorCategory.INVALID_VALUE,
                        target=ErrorTarget.TOOL_CALL_ACCURACY_EVALUATOR,
                    )
                eval_inputs.append({"query": query, "tool_call": tool_call, "tool_definition": tool_definition})
            else:
                raise EvaluationException(
                    message="Tool definition not found",
                    blame=ErrorBlame.USER_ERROR,
                    category=ErrorCategory.INVALID_VALUE,
                    target=ErrorTarget.TOOL_CALL_ACCURACY_EVALUATOR,
                )

        return eval_inputs

    @override
    async def _do_eval(self, eval_input: Dict) -> Dict[str, Union[float, str]]:  # type: ignore[override]
        """Do a relevance evaluation.

        :param eval_input: The input to the evaluator. Expected to contain
        whatever inputs are needed for the _flow method, including context
        and other fields depending on the child class.
        :type eval_input: Dict
        :return: The evaluation result.
        :rtype: Dict
        """
        llm_output = await self._flow(timeout=self._LLM_CALL_TIMEOUT, **eval_input)

        score = math.nan
        if llm_output:
            score, reason = parse_quality_evaluator_reason_score(llm_output, valid_score_range="[0-1]")
            if score >= 0 and score <= 1:
                return {
                    self._result_key: bool(float(score)),
                    f"{self._result_key}_reason": reason,
                    "tool_call_id" : eval_input.get("tool_call").get("tool_call_id"),
                }
        raise EvaluationException(
            message="Tool call accuracy evaluator: Invalid score returned from LLM.",
            blame=ErrorBlame.SYSTEM_ERROR,
            category=ErrorCategory.INVALID_VALUE,
            target=ErrorTarget.TOOL_CALL_ACCURACY_EVALUATOR,
        )

    async def _real_call(self, **kwargs):
        """The asynchronous call where real end-to-end evaluation logic is performed.

        :keyword kwargs: The inputs to evaluate.
        :type kwargs: Dict
        :return: The evaluation result.
        :rtype: Union[DoEvalResult[T_EvalValue], AggregateResult[T_EvalValue]]
        """
        # Convert inputs into list of evaluable inputs.
        eval_input_list = self._convert_kwargs_to_eval_input(**kwargs)
        if len(eval_input_list) == 0:
            return {self._AGGREGATE_RESULT_KEY: self._NOT_APPLICABLE_RESULT,
                    f"{self._AGGREGATE_RESULT_KEY}_result": self._NOT_APPLICABLE_RESULT,
                    f"{self._AGGREGATE_RESULT_KEY}_threshold": self.threshold,
                    f"{self._AGGREGATE_RESULT_KEY}_reason":
                        "No tool calls were made.",
                    "per_tool_call_details": []
                    }

        per_turn_results = []
        # Evaluate all inputs.
        for eval_input in eval_input_list:
            if self._is_applicable_tool(eval_input):
                per_turn_results.append(await self._do_eval(eval_input))
            else:
                per_turn_results.append(self._not_applicable_result(eval_input))

        return self._aggregate_results(per_turn_results=per_turn_results)

    def _is_applicable_tool(self, eval_input):
        """Determine if a given tool should be evaluated, since we only evaluate tools that
        have sufficient context available.

        :type eval_input: Dict
        :return: True if the tool call should be evaluated
        :rtype: bool
        """
        tool_definition = eval_input.get("tool_definition")
        if tool_definition is None or len(tool_definition) != 1:
            return False
        tool_type = tool_definition[0].get("type")
        if tool_type is None or tool_type != "function":
            return False
        return True

    def _not_applicable_result(self, eval_input):
        """Return a result indicating that the tool call is not applicable for evaluation.

        :param eval_input: The input to the evaluator.
        :type eval_input: Dict
        :return: A dictionary containing the result of the evaluation.
        :rtype: Dict[str, Union[str, float]]
        """
        return {
            f"{self._result_key}": self._NOT_APPLICABLE_RESULT,
            f"{self._result_key}_reason": "Tool call not supported for evaluation",
            "tool_call_id" : eval_input.get("tool_call").get("tool_call_id"),
        }

    def _aggregate_results(self, per_turn_results):
        """Aggregate the evaluation results of each conversation turn into a single result.

        Exact implementation might need to vary slightly depending on the results produced.
        Default behavior is to average the all number-based outputs.

        :param per_turn_results: List of evaluation results for each turn in the conversation.
        :type per_turn_results: List[Dict]
        :return: A dictionary containing aggregated results, with numeric metrics having their
        means as top-level values in the dictionary, and all original
        values (including non-numerics) located in under the "evaluation_per_turn" key,
        which each sub-key being a metric and each sub-value being a the list of that metric's
        per-turn values.
        :rtype: AggregateResult[T_EvalValue]
        """

        aggregated: Dict[str, Union[float, Dict[str, List[T_EvalValue]]]] = {}
        evaluation_per_turn: Dict[str, List[T_EvalValue]] = {}

        # Go over each turn, and rotate the results into a
        # metric: List[values] format for the evals_per_turn dictionary.

        num_evaluated = len([per_turn_result for per_turn_result in per_turn_results
                             if per_turn_result.get(self._result_key) != self._NOT_APPLICABLE_RESULT])
        if num_evaluated == 0:
            # None of the invoked tools were applicable, return not applicable result
            # (If a tool fails evaluation, we'll throw an exception)
            return {self._AGGREGATE_RESULT_KEY: self._NOT_APPLICABLE_RESULT,
                    f"{self._AGGREGATE_RESULT_KEY}_result": self._NOT_APPLICABLE_RESULT,
                    f"{self._AGGREGATE_RESULT_KEY}_threshold": self.threshold,
                    f"{self._AGGREGATE_RESULT_KEY}_reason":
                        "Tool call accuracy evaluation is not yet supported for the invoked tools.",
                    "per_tool_call_details": []
                    }
        # ignore not_applicable results, where the _result_key will be "not applicable"
        score = sum([per_turn_result.get(self._result_key) == True for per_turn_result in per_turn_results])/num_evaluated
        aggregated[self._AGGREGATE_RESULT_KEY] = score
        aggregated[f'{self._AGGREGATE_RESULT_KEY}_result'] = self._PASS_RESULT if score >= self.threshold else self._FAIL_RESULT
        aggregated[f'{self._AGGREGATE_RESULT_KEY}_threshold'] = self.threshold
        aggregated["per_tool_call_details"] = per_turn_results
        return aggregated

    @override
    def __call__(  # pylint: disable=docstring-missing-param
        self,
        *args,
        **kwargs,
    ):
        """
        Evaluate tool call accuracy. Accepts a query, tool definitions, and tool calls for evaluation.

        :keyword query: Query or Chat history up to the message that has the tool call being evaluated.
        :paramtype query: Union[str, List[dict]]
        :keyword tool_definitions: List of tool definitions whose calls are being evaluated.
        :paramtype tool_definitions: Union[dict, List[dict]]
        :keyword tool_calls: Optional List of tool calls to evaluate. If not provided response should be provided and should have
            tool call(s) in it.
        :paramtype tool_calls: Union[dict, List[dict]]
        :keyword response: Optional response to be evaluated alongside the tool calls.
            If provided all tool calls in response will be evaluated when tool_calls parameter is not provided.
            If provided and tool_calls parameter is provided, only the tool calls in tool_calls parameter will be evaluated.
                If response has extra tool calls they will not be evaluated, response will be used to extract any tool calls that are needed for evaluating a certain tool call.
            Recommended to provide it when there are tool calls that depend on output of a previous tool call.
        :paramtype response: Union[str, List[dict]]
        :return: The tool selection evaluation results.
        :rtype: Dict[str, Union[str, float]]
        """
        return super().__call__(*args, **kwargs)



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

import pytest

from azure.ai.evaluation import F1ScoreEvaluator
from azure.ai.evaluation import (
    AzureOpenAIGrader,
)
from azure.ai.evaluation import AzureOpenAIModelConfiguration


@pytest.fixture
def mock_aoai_model_config():
    return AzureOpenAIModelConfiguration(
            azure_deployment="...",
            azure_endpoint="...",
            api_key="...",
            api_version="...",
        )


@pytest.fixture
def mock_grader_config():
    return {}

def _get_file(name):
    """Get the file from the unittest data folder."""
    import os, pathlib
    data_path = os.path.join(pathlib.Path(__file__).parent.resolve(), "data")
    return os.path.join(data_path, name)

@pytest.fixture
def questions_file():
    return _get_file("questions.jsonl")

def simple_eval_function():
    return "123"

@pytest.mark.unittest
class TestRemoteEvaluationFeatures:

    def test_eval_name_mapping(self, mock_aoai_model_config, mock_grader_config):
        """
        Test that the name mapping works properly.
        TODO upgrade test to test EVERY KNOWN EVALUATOR.
        """
        # create f1 score eval
        f1_score_eval = F1ScoreEvaluator()
        # create aoai grader
        aoai_grader = AzureOpenAIGrader(model_config=mock_aoai_model_config, grader_config=mock_grader_config)
        from azure.ai.evaluation._evaluate._evaluate import _map_names_to_builtins
        from azure.ai.evaluation._eval_mapping import EVAL_CLASS_MAP

        evaluators = {
            "my_f1": f1_score_eval,
        }
        graders = {
            "my_aoai": aoai_grader,
        }
        result = _map_names_to_builtins(evaluators, graders)
        assert result is not None
        assert result["my_f1"] == EVAL_CLASS_MAP[F1ScoreEvaluator]
        assert result["my_aoai"] == AzureOpenAIGrader.id



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# To minimize relative imports in our evaluators, the scope of this package also includes anything
# that would have otherwise been a relative import scoped to single evaluator directories.

from . import constants
from .rai_service import evaluate_with_rai_service
from .utils import get_harm_severity_level
from .evaluation_onedp_client import EvaluationServiceOneDPClient
from .onedp.models import EvaluationUpload, EvaluationResult, RedTeamUpload, ResultType

__all__ = [
    "get_harm_severity_level",
    "evaluate_with_rai_service",
    "constants",
    "EvaluationServiceOneDPClient",
    "EvaluationResult",
    "EvaluationUpload",
    "RedTeamUpload",
    "ResultType",
]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from typing import Dict, TypeVar, Union, Optional

from typing_extensions import override

from azure.ai.evaluation._common.constants import (
    EvaluationMetrics,
    _InternalEvaluationMetrics,
    Tasks,
    _InternalAnnotationTasks,
)
from azure.ai.evaluation._common.rai_service import evaluate_with_rai_service, evaluate_with_rai_service_multimodal
from azure.ai.evaluation._common.utils import validate_azure_ai_project, is_onedp_project
from azure.ai.evaluation._exceptions import EvaluationException
from azure.ai.evaluation._common.utils import validate_conversation
from azure.ai.evaluation._constants import _AggregationType
from azure.core.credentials import TokenCredential

from . import EvaluatorBase

T = TypeVar("T")


class RaiServiceEvaluatorBase(EvaluatorBase[T]):
    """Base class for all evaluators that require the use of the Azure AI RAI service for evaluation.
    This includes content safety evaluators, protected material evaluators, and others. These evaluators
    are all assumed to be of the "query and response or conversation" input variety.

    :param eval_metric: The evaluation metric to be used for evaluation. This is used by the API call logic
        to specify which evaluation to perform.
    :type eval_metric: ~azure.ai.evaluation._common.constants.EvaluationMetrics
    :param eval_last_turn: If True, only the last turn of the conversation will be evaluated, and no
        aggregation will be performed. If False, all turns will be evaluated and the numeric results will be,
        aggregated. Per-turn results are still be available in the output via the "evaluation_per_turn" key
        when this occurs. Default is False, resulting full conversation evaluation and aggregation.
    :type eval_last_turn: bool
    :param conversation_aggregation_type: The type of aggregation to perform on the per-turn results of a conversation
        to produce a single result.
        Default is ~azure.ai.evaluation._AggregationType.MEAN.
    :type conversation_aggregation_type: ~azure.ai.evaluation._AggregationType
    :param threshold: The threshold for the evaluation. Default is 3.
    :type threshold: Optional[int]
    :param _higher_is_better: If True, higher scores are better. Default is True.
    :type _higher_is_better: Optional[bool]
    """

    @override
    def __init__(
        self,
        eval_metric: Union[EvaluationMetrics, _InternalEvaluationMetrics],
        azure_ai_project: Union[dict, str],
        credential: TokenCredential,
        eval_last_turn: bool = False,
        conversation_aggregation_type: _AggregationType = _AggregationType.MEAN,
        threshold: int = 3,
        _higher_is_better: Optional[bool] = False,
    ):
        super().__init__(eval_last_turn=eval_last_turn, conversation_aggregation_type=conversation_aggregation_type, threshold=threshold, _higher_is_better=_higher_is_better)
        self._eval_metric = eval_metric
        self._azure_ai_project = validate_azure_ai_project(azure_ai_project)
        self._credential = credential
        self._threshold = threshold
        self._higher_is_better = _higher_is_better

    @override
    def __call__(  # pylint: disable=docstring-missing-param
        self,
        *args,
        **kwargs,
    ):
        """Evaluate either a query and response or a conversation. Must supply either a query AND response,
        or a conversation, but not both.

        :keyword query: The query to evaluate.
        :paramtype query: Optional[str]
        :keyword response: The response to evaluate.
        :paramtype response: Optional[str]
        :keyword conversation: The conversation to evaluate. Expected to contain a list of conversation turns under the
            key "messages", and potentially a global context under the key "context". Conversation turns are expected
            to be dictionaries with keys "content", "role", and possibly "context".
        :paramtype conversation: Optional[~azure.ai.evaluation.Conversation]
        :rtype: Union[Dict[str, T], Dict[str, Union[float, Dict[str, List[T]]]]]
        """
        return super().__call__(*args, **kwargs)

    @override
    async def _do_eval(self, eval_input: Dict) -> Dict[str, T]:
        """Perform the evaluation using the Azure AI RAI service.
        The exact evaluation performed is determined by the evaluation metric supplied
        by the child class initializer.

        :param eval_input: The input to the evaluation function.
        :type eval_input: Dict
        :return: The evaluation result.
        :rtype: Dict
        """
        if "query" in eval_input and "response" in eval_input:
            return await self._evaluate_query_response(eval_input)

        conversation = eval_input.get("conversation", None)
        return await self._evaluate_conversation(conversation)

    async def _evaluate_conversation(self, conversation: Dict) -> Dict[str, T]:
        """
        Evaluates content according to this evaluator's metric.
        :keyword conversation: The conversation contains list of messages to be evaluated.
            Each message should have "role" and "content" keys.

        :param conversation: The conversation to evaluate.
        :type conversation: ~azure.ai.evaluation.Conversation
        :return: The evaluation score computation based on the Content Safety metric (self.metric).
        :rtype: Dict[str, Union[float, str]]
        """
        # validate inputs
        validate_conversation(conversation)
        messages = conversation["messages"]
        # Run score computation based on supplied metric.
        result = await evaluate_with_rai_service_multimodal(
            messages=messages,
            metric_name=self._eval_metric,
            project_scope=self._azure_ai_project,
            credential=self._credential,
        )
        return result

    async def _evaluate_query_response(self, eval_input: Dict) -> Dict[str, T]:
        query = str(eval_input.get("query", None))
        response = str(eval_input.get("response", None))
        if query is None or response is None:
            raise EvaluationException(
                message="Not implemented",
                internal_message=(
                    "Reached query/response evaluation without supplying query or response."
                    + " This should have failed earlier."
                ),
            )
        input_data = {"query": query, "response": response}

        if "context" in self._singleton_inputs:
            context = eval_input.get("context", None)
            if context is None:
                raise EvaluationException(
                    message="Not implemented",
                    internal_message=(
                        "Attempted context-based evaluation without supplying context."
                        + " This should have failed earlier."
                    ),
                )
            input_data["context"] = context

        return await evaluate_with_rai_service(  # type: ignore
            metric_name=self._eval_metric,
            data=input_data,
            project_scope=self._azure_ai_project,
            credential=self._credential,
            annotation_task=self._get_task(),
            evaluator_name=self.__class__.__name__,
        )

    def _get_task(self):
        """Get the annotation task for the current evaluation metric.
        The annotation task is used by the RAI service script to determine a the message format
        of the API call, and how the output is processed, among other things.

        :return: The annotation task for the evaluator's self._eval_metric value.
        :rtype: ~azure.ai.evaluation._common.constants.Tasks

        """
        if self._eval_metric == EvaluationMetrics.GROUNDEDNESS:
            return Tasks.GROUNDEDNESS
        if self._eval_metric == EvaluationMetrics.XPIA:
            return Tasks.XPIA
        if self._eval_metric == _InternalEvaluationMetrics.ECI:
            return _InternalAnnotationTasks.ECI
        if self._eval_metric == EvaluationMetrics.PROTECTED_MATERIAL:
            return Tasks.PROTECTED_MATERIAL
        if self._eval_metric == EvaluationMetrics.CODE_VULNERABILITY:
            return Tasks.CODE_VULNERABILITY
        if self._eval_metric == EvaluationMetrics.UNGROUNDED_ATTRIBUTES:
            return Tasks.UNGROUNDED_ATTRIBUTES
        return Tasks.CONTENT_HARM



"""
Unit tests for red_team.utils.constants module.
"""

import pytest
try:
    import pyrit
    has_pyrit = True
except ImportError:
    has_pyrit = False

if has_pyrit:
    from azure.ai.evaluation.red_team._utils.constants import (
        BASELINE_IDENTIFIER, DATA_EXT, RESULTS_EXT,
        ATTACK_STRATEGY_COMPLEXITY_MAP, RISK_CATEGORY_EVALUATOR_MAP,
        INTERNAL_TASK_TIMEOUT, TASK_STATUS
    )
    from azure.ai.evaluation.red_team._attack_strategy import AttackStrategy
    from azure.ai.evaluation.red_team._attack_objective_generator import RiskCategory
from azure.ai.evaluation import ViolenceEvaluator, HateUnfairnessEvaluator, SexualEvaluator, SelfHarmEvaluator


@pytest.mark.unittest
@pytest.mark.skipif(not has_pyrit, reason="redteam extra is not installed")
class TestBasicConstants:
    """Test basic constants are defined correctly."""

    def test_basic_constants(self):
        """Verify basic constants are defined correctly."""
        assert BASELINE_IDENTIFIER == "baseline"
        assert DATA_EXT == ".jsonl"
        assert RESULTS_EXT == ".json"
        assert INTERNAL_TASK_TIMEOUT == 120


@pytest.mark.unittest
@pytest.mark.skipif(not has_pyrit, reason="redteam extra is not installed")
class TestTaskStatusConstants:
    """Test task status constants are defined correctly."""
    
    def test_task_status_constants(self):
        """Verify task status constants are defined correctly."""
        assert TASK_STATUS["PENDING"] == "pending"
        assert TASK_STATUS["RUNNING"] == "running"
        assert TASK_STATUS["COMPLETED"] == "completed"
        assert TASK_STATUS["FAILED"] == "failed"
        assert TASK_STATUS["TIMEOUT"] == "timeout"


@pytest.mark.unittest
@pytest.mark.skipif(not has_pyrit, reason="redteam extra is not installed")
class TestRiskCategoryEvaluatorMap:
    """Test risk category evaluator mapping."""
    
    def test_risk_category_evaluator_map(self):
        """Verify risk category evaluator map includes all risk categories."""
        assert RiskCategory.Violence in RISK_CATEGORY_EVALUATOR_MAP
        assert RISK_CATEGORY_EVALUATOR_MAP[RiskCategory.Violence] == ViolenceEvaluator
        
        assert RiskCategory.HateUnfairness in RISK_CATEGORY_EVALUATOR_MAP
        assert RISK_CATEGORY_EVALUATOR_MAP[RiskCategory.HateUnfairness] == HateUnfairnessEvaluator
        
        assert RiskCategory.Sexual in RISK_CATEGORY_EVALUATOR_MAP
        assert RISK_CATEGORY_EVALUATOR_MAP[RiskCategory.Sexual] == SexualEvaluator
        
        assert RiskCategory.SelfHarm in RISK_CATEGORY_EVALUATOR_MAP
        assert RISK_CATEGORY_EVALUATOR_MAP[RiskCategory.SelfHarm] == SelfHarmEvaluator


# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from enum import Enum
from typing import List
from azure.ai.evaluation._common._experimental import experimental


@experimental
class AttackStrategy(Enum):
    """Strategies for attacks."""
    EASY = "easy"
    MODERATE = "moderate"
    DIFFICULT = "difficult"
    AnsiAttack = "ansi_attack"
    AsciiArt = "ascii_art"
    AsciiSmuggler = "ascii_smuggler"
    Atbash = "atbash"
    Base64 = "base64"
    Binary = "binary"
    Caesar = "caesar"
    CharacterSpace = "character_space"
    CharSwap = "char_swap"
    Diacritic = "diacritic"
    Flip = "flip"
    Leetspeak = "leetspeak"
    Morse = "morse"
    ROT13 = "rot13"
    SuffixAppend = "suffix_append"
    StringJoin = "string_join"
    Tense = "tense"
    UnicodeConfusable = "unicode_confusable"
    UnicodeSubstitution = "unicode_substitution"
    Url = "url"
    Baseline = "baseline"
    Jailbreak = "jailbreak"

    @classmethod
    def Compose(cls, items: List["AttackStrategy"]) -> List["AttackStrategy"]:
        for item in items:
            if not isinstance(item, cls):
                raise ValueError("All items must be instances of AttackStrategy")
        if len(items) > 2: 
            raise ValueError("Composed strategies must have at most 2 items")
        return items



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Any
from typing_extensions import TypeAlias


try:
    from promptflow.core._errors import MissingRequiredPackage as _MissingRequiredPackage
except ImportError:
    from azure.ai.evaluation._exceptions import ErrorBlame, ErrorCategory, ErrorTarget, EvaluationException

    class _MissingRequiredPackage(EvaluationException):
        """Raised when a required package is missing.

        :param message: A message describing the error. This is the error message the user will see.
        :type message: str
        """

        def __init__(self, message: str, **kwargs: Any):
            kwargs.setdefault("category", ErrorCategory.MISSING_PACKAGE)
            kwargs.setdefault("blame", ErrorBlame.SYSTEM_ERROR)
            kwargs.setdefault("target", ErrorTarget.EVALUATE)
            kwargs.setdefault("internal_message", "Missing required package.")
            super().__init__(message=message, **kwargs)


MissingRequiredPackage: TypeAlias = _MissingRequiredPackage



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from ._evaluate._evaluate import evaluate
from ._evaluators._bleu import BleuScoreEvaluator
from ._evaluators._coherence import CoherenceEvaluator
from ._evaluators._content_safety import (
    ContentSafetyEvaluator,
    HateUnfairnessEvaluator,
    SelfHarmEvaluator,
    SexualEvaluator,
    ViolenceEvaluator,
)
from ._evaluators._f1_score import F1ScoreEvaluator
from ._evaluators._fluency import FluencyEvaluator
from ._evaluators._gleu import GleuScoreEvaluator
from ._evaluators._groundedness import GroundednessEvaluator
from ._evaluators._service_groundedness import GroundednessProEvaluator
from ._evaluators._intent_resolution import IntentResolutionEvaluator
from ._evaluators._meteor import MeteorScoreEvaluator
from ._evaluators._protected_material import ProtectedMaterialEvaluator
from ._evaluators._qa import QAEvaluator
from ._evaluators._response_completeness import ResponseCompletenessEvaluator
from ._evaluators._task_adherence import TaskAdherenceEvaluator
from ._evaluators._relevance import RelevanceEvaluator
from ._evaluators._retrieval import RetrievalEvaluator
from ._evaluators._rouge import RougeScoreEvaluator, RougeType
from ._evaluators._similarity import SimilarityEvaluator
from ._evaluators._xpia import IndirectAttackEvaluator
from ._evaluators._code_vulnerability import CodeVulnerabilityEvaluator
from ._evaluators._ungrounded_attributes import UngroundedAttributesEvaluator
from ._evaluators._tool_call_accuracy import ToolCallAccuracyEvaluator
from ._evaluators._document_retrieval import DocumentRetrievalEvaluator
from ._model_configurations import (
    AzureAIProject,
    AzureOpenAIModelConfiguration,
    Conversation,
    EvaluationResult,
    EvaluatorConfig,
    Message,
    OpenAIModelConfiguration,
)
from ._aoai.aoai_grader import AzureOpenAIGrader
from ._aoai.label_grader import AzureOpenAILabelGrader
from ._aoai.string_check_grader import AzureOpenAIStringCheckGrader
from ._aoai.text_similarity_grader import AzureOpenAITextSimilarityGrader


_patch_all = []

# The converter from the AI service to the evaluator schema requires a dependency on
# ai.projects, but we also don't want to force users installing ai.evaluations to pull
# in ai.projects. So we only import it if it's available and the user has ai.projects.
try:
    from ._converters._ai_services import AIAgentConverter
    _patch_all.append("AIAgentConverter")
except ImportError:
    print("[INFO] Could not import AIAgentConverter. Please install the dependency with `pip install azure-ai-projects`.")


__all__ = [
    "evaluate",
    "CoherenceEvaluator",
    "F1ScoreEvaluator",
    "FluencyEvaluator",
    "GroundednessEvaluator",
    "GroundednessProEvaluator",
    "ResponseCompletenessEvaluator",
    "TaskAdherenceEvaluator",
    "IntentResolutionEvaluator",
    "RelevanceEvaluator",
    "SimilarityEvaluator",
    "QAEvaluator",
    "ViolenceEvaluator",
    "SexualEvaluator",
    "SelfHarmEvaluator",
    "HateUnfairnessEvaluator",
    "ContentSafetyEvaluator",
    "IndirectAttackEvaluator",
    "BleuScoreEvaluator",
    "GleuScoreEvaluator",
    "MeteorScoreEvaluator",
    "RetrievalEvaluator",
    "RougeScoreEvaluator",
    "RougeType",
    "ProtectedMaterialEvaluator",
    "AzureAIProject",
    "AzureOpenAIModelConfiguration",
    "OpenAIModelConfiguration",
    "EvaluatorConfig",
    "Conversation",
    "Message",
    "EvaluationResult",
    "CodeVulnerabilityEvaluator",
    "UngroundedAttributesEvaluator",
    "ToolCallAccuracyEvaluator",
    "AzureOpenAIGrader",
    "AzureOpenAILabelGrader",
    "AzureOpenAIStringCheckGrader",
    "AzureOpenAITextSimilarityGrader",
]

__all__.extend([p for p in _patch_all if p not in __all__])